# -*- coding: utf-8 -*-
"""Components for processes in a depot."""
import math
import warnings
from abc import ABC, abstractmethod
from enum import auto, Enum
from warnings import warn

import simpy
from eflips.helperFunctions import flexprint
from eflips.settings import globalConstants

import eflips
from eflips.depot.evaluation import (
    BatteryLog,
    ChargeStart,
    FullyCharged,
    ProcessFinished,
)
from eflips.depot.filters import VehicleFilter


class ProcessStatus(Enum):
    """Status codes for depot processes."""

    NOT_STARTED = auto()  # process hasn't started yet
    WAITING = auto()  # process is waiting for required resources
    IN_PROGRESS = auto()  # process is being executed
    COMPLETED = auto()  # process successfully finished
    CANCELLED = auto()  # process was interrupted with resume=False


class EstimateValue(Enum):
    """Values for process completion time estimates in addition to an int.

    value.
    """

    UNKNOWN = auto()  # process not completed but an estimate is not possible
    COMPLETED = auto()  # process completed or no process scheduled


class BaseDepotProcess(ABC):
    """Base class for processes in a depot.

    Core functionalities are:
     - wrapping the execution of a simpy process and handling of required
        resources,
     - providing the option to resume the process after interruption.

    Implements __call__ to make an instance directly callable through
    env.process(BaseDepotProcessSubclass(...)).

    An instance is not reusable, meaning a new object should be instantiated
    for every execution (except resumption).

    A new subclass must be added to configuration.map_process_cls_refs.

    Parameters:
    ID: [str] unique name of the process, such as "serve_default". Subclasses
        instances may use the same ID.
    dur: [int or None] Duration of the execution in seconds. Can be left None
        upon instantiation (e.g. if the duration is calculated directly before
        execution).
    required_resources: [list] of DepotResource or subclass objects. May be
        empty. Before the actual process begins, the execution waits for
        requests to all of those resources to succeed.
    resume: [bool] If True, an interrupted process is resumed for the remaining
        duration.
    priority: [int] priority* for requests to resources in required_resources.
    recall_priority: [int] priority* for re-requests to resources after
        interruption.
    *Note that custom values for priority and recall_priority should be higher
        than simpyExt.ResourceSwitch.priority for Resourceswitch to keep
        its priority (lower value means higher priority).
    preempt: [bool] Argument for requests to resources in required_resources.
        See class DepotResourceRequest and SimPy-doc for details.
    dur_predefined: [bool] True if parameter *dur* is defined before (re)start
        of this process. Set to False if the duration is calculated after
        restarting the process (such as for Charge).

    Attributes:
    starts: [list] of points of time [int] when the process has (re-)started.
    proc: [simpy.Process] processing self._pem.
    requests: [list] of DepotResourceRequest objects for the current process
        execution.
    status: [ProcessStatus] member.
    finished: [simpy.Event] succeeds when the process is completed or cancelled
        and won't be resumed, resulting in self.__call__ to proceed.
    etc: [int or EstimateValue] estimated simulation time of completion. See
        method self._estimate_time_of_completion for details.
    """

    @abstractmethod
    def __init__(
        self,
        env,
        ID,
        dur=None,
        required_resources=None,
        resume=True,
        priority=0,
        recall_priority=-1,
        preempt=True,
        dur_predefined=True,
    ):
        self.env = env
        self.ID = ID
        self.dur = dur
        self.dur_orig = dur

        self.required_resources = (
            required_resources if required_resources is not None else []
        )
        self.resume = resume
        self.priority = priority
        self.recall_priority = recall_priority
        self.preempt = preempt

        self.dur_predefined = dur_predefined

        self.recall_count = 0
        self.starts = []
        self.ends = []
        self.proc = None
        self.requests = []
        self._status = ProcessStatus.NOT_STARTED
        self.finished = env.event()
        self.etc = EstimateValue.UNKNOWN

        if (
            required_resources is not None
            and not preempt
            and len(required_resources) > 1
        ):
            warn(
                'BaseDepotProcess parameter "preempt" should not be False '
                'while there is more than 1 resource in "required_resources" '
                "unless a resource-locking case can be precluded. See"
                " DepotResourceRequest.preempt for more info."
            )

    @property
    def resIDs(self):
        """Return a list of IDs of self.required_resources."""
        return [res.ID for res in self.required_resources]

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        self._status = value
        self._estimate_time_of_completion()

    def _estimate_time_of_completion(self):
        """Calculate self.etc based on the current status."""
        if (
            self.status is ProcessStatus.NOT_STARTED
            or self.status is ProcessStatus.WAITING
        ):
            self.etc = EstimateValue.UNKNOWN

        elif self.status is ProcessStatus.IN_PROGRESS:
            if self.dur is None:
                raise ValueError(
                    "Process %s attribute 'dur' must be set "
                    "before process execution." % self.ID
                )
            else:
                self.etc = self.env.now + self.dur

        elif (
            self.status is ProcessStatus.COMPLETED
            or self.status is ProcessStatus.CANCELLED
        ):
            self.etc = EstimateValue.COMPLETED

    @abstractmethod
    def _action(self, *args, **kwargs):
        """Generator function that characterizes the process.

        Subclasses must
        implement this method. It must catch the exception simpy.Interrupt and
        let it pass or implement custom follow-ups. If this method yields a
        timeout, the duration must be set as self.dur before the yield
        statement due to the determination of self.etc.
        """

    @staticmethod
    @abstractmethod
    def estimate_duration(*args, **kwargs):
        """Return an estimate of the process duration in seconds based on given.

        parameters as as if it could start immediately. Must be int, not float.
        Possible waiting time for resources are not considered. Therefore the
        estimate is reliable only if the process is started immediately. Static
        to be callable before instantiation.
        """

    def _pre(self, *args, **kwargs):
        """Space for actions before calling self._pem that may not be possible.

        upon init.
        """
        pass

    def _post(self, *args, **kwargs):
        """Space for actions after finishing self._pem."""
        pass

    def __call__(self, *args, **kwargs):
        """Generator that starts the execution of this process.

        Note that all parameters required by self._action() have to be passed
        as keyword arguments.
        """
        self._pre(*args, **kwargs)

        try:
            self.proc = self.env.process(self._pem(*args, **kwargs))
            yield self.finished

        except simpy.Interrupt:
            raise RuntimeError(
                "BaseDepotProcess must be interrupted through "
                "BaseDepotProcess.proc.interrupt() instead of "
                "interrupting __call__."
            )

        self._post(*args, **kwargs)

    def _pem(self, recall=False, *args, **kwargs):
        """Process execution method (pem) that wraps waiting for required.

        resources (if any) and self._action in a single generator. Is called
        again for the remaining duration after interruption if self.resume is
        True.

        recall: [bool] False for the first call to this method, which uses
            self.priority and self.dur. For a recall, self.recall_priority and
            overwritten self.dur are used.
        """
        prio = self.recall_priority if recall else self.priority
        self.requests = []
        action_proc = None

        try:
            # Wait for required resources
            self._request_resources(prio)
            self.status = ProcessStatus.WAITING
            yield self.env.all_of(self.requests)

            if hasattr(self, "vehicle") and self.vehicle is not None:
                flexprint(
                    'Process "%s" for vehicle %s got resources.'
                    % (self.ID, self.vehicle.ID),
                    env=self.env,
                    switch="processes",
                )

            start_time = self.env.now
            self.proc.action_start = start_time
            self.starts.append(start_time)
            # Call actual process method
            action_proc = self.env.process(self._action(*args, **kwargs))
            # Step back in event schedule to let action_proc execute until the
            # first yield statement, i.e. calculate self.dur if unknown
            # beforehand
            yield self.env.timeout(0)
            self.status = ProcessStatus.IN_PROGRESS
            yield action_proc

            if hasattr(self, "vehicle") and self.vehicle is not None:
                flexprint(
                    'Process "%s" for vehicle %s End of pem reached.'
                    % (self.ID, self.vehicle.ID),
                    env=self.env,
                    switch="processes",
                )

            self._cleanup()
            self.status = ProcessStatus.COMPLETED
            if hasattr(self, "vehicle") and self.vehicle is not None:
                self.vehicle.dwd.active_processes.remove(self)
            self.finished.succeed()

        except simpy.Interrupt:
            if hasattr(self, "vehicle") and self.vehicle is not None:
                flexprint(
                    'Process "%s" for vehicle %s interrupted.'
                    % (self.ID, self.vehicle.ID),
                    env=self.env,
                    switch="processes",
                )

            if not self.resume:
                if hasattr(
                    self.vehicle.dwd.current_area, "items"
                ):  # Needed because of Precondition, special cancel time
                    items = self.vehicle.dwd.current_area.items
                else:
                    items = None
                flexprint(
                    'Process "%s" for vehicle %s interrupted. area. %s, area items: %s. resume: %s'
                    % (
                        self.ID,
                        self.vehicle.ID,
                        self.vehicle.dwd.current_area,
                        items,
                        self.resume,
                    ),
                    env=self.env,
                    switch="departure_before_fully_charged_3",
                )

            # Interrupt process action_proc
            if action_proc is not None and not action_proc.triggered:
                action_proc.interrupt()

            if self.resume:  # Process may be resumed
                self._cleanup()

                if self.dur_predefined:
                    # Process requires pre-definition of the (remaining)
                    # duration
                    usage_time = (
                        self.env.now - self.proc.start
                        if hasattr(self.proc, "start")
                        else 0
                    )
                    remaining_time = self.dur - usage_time

                    if remaining_time > 0:
                        self.recall_count += 1

                        # Update dur for the recall
                        self.dur = remaining_time

                        # Call process again
                        if hasattr(self, "vehicle") and self.vehicle is not None:
                            flexprint(
                                'Process "%s" for vehicle %s will be resumed with dur=%s.'
                                % (self.ID, self.vehicle.ID, remaining_time),
                                env=self.env,
                                switch="processes",
                            )

                        self.proc = self.env.process(
                            self._pem(recall=True, *args, **kwargs)
                        )

                else:
                    # Continue without dur (for processes such as Charge).
                    # Call process again
                    if hasattr(self, "vehicle") and self.vehicle is not None:
                        flexprint(
                            'Process "%s" for vehicle %s will be resumed.'
                            % (self.ID, self.vehicle.ID),
                            env=self.env,
                            switch="processes",
                        )
                    self.proc = self.env.process(
                        self._pem(recall=True, *args, **kwargs)
                    )

            else:
                if hasattr(self, "vehicle") and self.vehicle is not None:
                    flexprint(
                        "Process '%s' for vehicle %s won't be resumed."
                        % (self.ID, self.vehicle.ID),
                        env=self.env,
                        switch="processes",
                    )
                self._cleanup_after_cancel()

    def _request_resources(self, prio):
        """Issue requests to required resources, if any."""
        if self.required_resources:
            resIDs = self.resIDs

            if hasattr(self, "vehicle") and self.vehicle is not None:
                flexprint(
                    'Process "%s" for vehicle %s requesting resources %s with prio=%s; dur=%s.'
                    % (self.ID, self.vehicle.ID, resIDs, prio, self.dur),
                    env=self.env,
                    switch="processes",
                )

            self.requests = [
                res.request(caller=self, priority=prio, preempt=self.preempt)
                for res in self.required_resources
            ]

    def _release_resources(self):
        """Release occupied resource slots and open requests."""
        for req in self.requests:
            if req.triggered:
                req.resource.release(req)
            else:
                req.cancel()

    def _cleanup(self):
        """Actions after process execution regardless of its success."""
        if hasattr(self, "vehicle") and self.vehicle is not None:
            flexprint(
                'Process "%s" for vehicle %s cleanup called'
                % (self.ID, self.vehicle.ID),
                env=self.env,
                switch="processes",
            )
        # Release occupied resource slots and open requests
        self._release_resources()

        # For stats
        if hasattr(self.proc, "action_start"):
            self.ends.append(self.env.now)

        if hasattr(self, "vehicle") and self.vehicle is not None:
            flexprint(
                'Process "%s" for vehicle %s cleaned up' % (self.ID, self.vehicle.ID),
                env=self.env,
                switch="processes",
            )

    def cancel(self):
        """Immediately stop the current process execution.

        No restart is
        scheduled, even if self.resume was set to True before.
        """
        if hasattr(
            self.vehicle.dwd.current_area, "items"
        ):  # Needed because of Precondition, special cancel time
            items = self.vehicle.dwd.current_area.items
        else:
            items = None
        flexprint(
            "Cancelling %s for vehicle %s. soc=%s, area: %s, area items: %s"
            % (
                self,
                self.vehicle.ID,
                self.vehicle.battery.soc,
                self.vehicle.dwd.current_area,
                items,
            ),
            env=self.env,
            switch="departure_before_fully_charged_3",
        )

        self.resume = False
        self.proc.interrupt()

    def _cleanup_after_cancel(self):
        """Do general and cancel-specific cleanup."""
        if hasattr(self, "vehicle") and self.vehicle is not None:
            flexprint(
                "Process '%s' for vehicle %s cancelled." % (self.ID, self.vehicle.ID),
                env=self.env,
                switch="processes",
            )
        self._cleanup()
        self.status = ProcessStatus.CANCELLED
        if hasattr(self, "vehicle") and self.vehicle is not None:
            self.vehicle.dwd.active_processes.remove(self)
        self.finished.succeed()

    def interrupt(self):
        """Immediately stop the current process execution.

        If self.resume is
        True, a restart is scheduled and may be successful immediately if
        required resources are available. If self.resume is False, same effect
        as self.cancel.
        """
        self.proc.interrupt()


class VehicleProcess(BaseDepotProcess, ABC):
    """Base class for processes specific for vehicles at areas.

    Parameters:
    ismandatory: [bool] True if the process is mandatory for vehicles. False if
        the process is optional. If optional, set vehicle_filter for further
        decision.
    vehicle_filter: [VehicleFilter] object for determining if the process
        should be executed for a vehicle on the area. If left None, all
        vehicles are permitted.
    cancellable_for_dispatch: [bool]: if True, the process may be cancelled for
        dispatching the vehicle.

    Attributes:
    request_immediately: [bool] if True, a vehicle will request this process to
        start immediately after entering the area this process is available at.
        There might still be waiting time due to resource requirements.
    """

    request_immediately = True

    def __init__(
        self,
        env,
        ID,
        dur,
        ismandatory=True,
        vehicle_filter=None,
        required_resources=None,
        resume=True,
        priority=0,
        recall_priority=-1,
        preempt=True,
        cancellable_for_dispatch=False,
        dur_predefined=True,
    ):
        super(VehicleProcess, self).__init__(
            env,
            ID,
            dur,
            required_resources,
            resume,
            priority,
            recall_priority,
            preempt,
            dur_predefined,
        )

        self.vehicle = None
        self.ismandatory = ismandatory
        self.vehicle_filter = (
            vehicle_filter if vehicle_filter is not None else VehicleFilter()
        )

        self.cancellable_for_dispatch = cancellable_for_dispatch
        if cancellable_for_dispatch:
            self.finished.callbacks.append(self.notify_assignment)

    def _pre(self, *args, **kwargs):
        self.vehicle = kwargs["vehicle"]
        # flexprint('in _pre of %s for vehicle %s. slot: %s'
        #         % (self.ID, self.vehicle.ID, self.vehicle.dwd.current_slot),
        #         env=self.env, switch='objID', objID=self.vehicle.ID)
        self.vehicle.dwd.active_processes.append(self)

        if (
            hasattr(self.vehicle, "logger")
            and globalConstants["general"]["LOG_ATTRIBUTES"]
            and globalConstants["general"]["LOG_SPECIFIC_STEPS"]
        ):
            self.vehicle.logger.steplog()

    def _post(self, *args, **kwargs):
        if globalConstants["depot"]["log_cm_data"]:
            # Use home_depot in case the vehicle left the depot before
            # finishing the process (not robust for simulations with multiple
            # depots)
            self.vehicle.dwd.home_depot.evaluation.cm_report.log(
                ProcessFinished(self.env, self.vehicle)
            )
        # flexprint(
        #     'in _post of %s for vehicle %s. slot: %s'
        #     % (self.ID, self.vehicle.ID, self.vehicle.dwd.current_slot),
        #     env=self.env, switch='objID', objID=self.vehicle.ID)

    def notify_assignment(self, *args):
        """Trigger dispatch process if there is a relevant update."""
        if (
            self.vehicle is not None
            and self.vehicle.dwd.current_area.issink
            and self.vehicle.dwd.active_processes
            and all(
                [
                    proc.cancellable_for_dispatch
                    for proc in self.vehicle.dwd.active_processes
                ]
            )
        ):
            flexprint(
                "NOTIFYING ASSIGNMENT AFTER %s END for vehicle %s"
                % (self.ID, self.vehicle.ID),
                env=self.env,
            )

            self.vehicle.dwd.current_depot.depot_control.trigger_dispatch()


class Serve(VehicleProcess):
    """Process of serving a vehicle such as cleaning."""

    def __init__(
        self,
        env,
        ID,
        dur,
        ismandatory=True,
        vehicle_filter=None,
        required_resources=None,
        resume=True,
        priority=0,
        recall_priority=-1,
        preempt=True,
        cancellable_for_dispatch=False,
    ):
        super(Serve, self).__init__(
            env,
            ID,
            dur,
            ismandatory,
            vehicle_filter,
            required_resources,
            resume,
            priority,
            recall_priority,
            preempt,
            cancellable_for_dispatch,
        )

    def _action(self, vehicle):
        try:
            flexprint(
                "\t%s started service." % (vehicle.ID),
                env=self.env,
                switch="operations",
            )
            flexprint(
                "\t%s started service." % (vehicle.ID), env=self.env, switch="processes"
            )
            yield self.env.timeout(self.dur)
            vehicle.dwd.service_need = False
            vehicle.dwd.t_lastServed = self.env.now
            flexprint(
                "\t%s completed service." % (vehicle.ID),
                env=self.env,
                switch="processes",
            )

        except simpy.Interrupt:
            flexprint("action interrupted", env=self.env, switch="processes")
            pass

    @staticmethod
    def estimate_duration(dur, *args, **kwargs):
        return dur


class ChargeAbstract(VehicleProcess, ABC):
    """Base class for charging of vehicles at a depot area.

    Parameters:
    dur: is None because the duration is set during charging.
    required_resources: doesn't need to contain the charging interface because
        it's detected automatically.
    efficiency: [int or float] efficiency of charging process. Value higher
        than 0, max 1.

    Attributes:
    last_update: [int] time of last call to self.update_battery.
    """

    @abstractmethod
    def __init__(
        self,
        env,
        ID,
        ismandatory=True,
        vehicle_filter=None,
        required_resources=None,
        resume=True,
        priority=0,
        recall_priority=-1,
        preempt=True,
        cancellable_for_dispatch=False,
        efficiency=1,
        *args,
        **kwargs
    ):
        if required_resources is not None:
            raise ValueError(
                "Parameter 'required_resources' is deactivated for process "
                "ChargeAbstract. The charging interface is detected "
                "automatically. Remove this statement when adding other "
                "requirements."
            )

        super(ChargeAbstract, self).__init__(
            env,
            ID,
            None,
            ismandatory,
            vehicle_filter,
            required_resources,
            resume,
            priority,
            recall_priority,
            preempt,
            cancellable_for_dispatch,
            dur_predefined=False,
        )

        self.efficiency = efficiency
        self.last_update = 0
        self.energy = 0  # total Energy of charging process
        self.charging_interface = None  # is set upon call

    def _pre(self, *args, **kwargs):
        super()._pre(*args, **kwargs)
        self.charging_interface = self.vehicle.dwd.current_charging_interface

    def _request_resources(self, prio):
        """Request the charging interface at the vehicle's current slot."""

        flexprint(
            'Process "%s" for vehicle %s requesting resource %s'
            " with prio=%s; dur=%s. Index: %d, slot_no: %d"
            % (
                self.ID,
                self.vehicle.ID,
                self.charging_interface.ID,
                prio,
                self.dur,
                self.vehicle.dwd.current_area.items.index(self.vehicle),
                self.vehicle.dwd.current_slot,
            ),
            env=self.env,
            switch="processes",
        )

        self.requests = [
            self.charging_interface.request(
                caller=self, priority=prio, preempt=self.preempt
            )
        ]

    @classmethod
    def get_chargedata(cls, vehicle):
        """Get data for charging process of class *cls*.

        Not robust if there
        are multiple data sets using *cls* with different parameters and
        vehicle filters that can return True for the same vehicle.
        """
        return next(
            (
                procdata
                for procdata in vehicle.dwd.current_depot.processes.values()
                if procdata["type"] is cls
                and procdata["kwargs"]["vehicle_filter"](vehicle)
            ),
            None,
        )

    def update_battery(self, event_name, amount=None):
        """Update the energy level of self.vehicle.battery.

        Can be called
        during the execution of the process to provide an interim update.

        Parameters:
        event_name: [str] for BatteryLog
        amount: [None or float] amount of energy to add to the battery. Must be
            passed if amount might be so small that t_elapsed would be 0 (i.e.
            if called by Charge(sub-)-classes). Leave None for update requests
            from outside the process.
        """
        self.last_update = self.env.now

        if amount is None and (
            self.last_update == self.env.now or self.starts[-1] == self.env.now
        ):
            # Skip if the call is not for putting and either the process just
            # started (possibly waiting for resources) or just updated
            return

        if amount is None:
            t_base = max(self.last_update, self.starts[-1])
            t_elapsed = self.env.now - t_base
            amount = t_elapsed / 3600 * self.power * self.efficiency
            rest = round(
                self.vehicle.battery.energy_real * self.soc_target
                - self.vehicle.battery.energy,
                14,
            )
            if rest < amount:
                # Process was about to finish in this second. Reduce amount to
                # rest to avoid soc > soc_target and provide an early update
                amount = rest
                event_name = "charge_end"

        self.vehicle.battery.put(amount)
        self.energy += amount
        self.vehicle.battery_logs.append(
            BatteryLog(self.env.now, self.vehicle, event_name)
        )


class Charge(ChargeAbstract):
    """Process of charging a vehicle's battery.

    Charging is constant during the
    whole charging process with maximum power provided by the charging
    interface.

    Parameters:
    soc_target: [int or float or str] the charging process will stop at this
        soc. Must be 0 < soc_target <= 1 or 'soc_max'. For soc_max,
        vehicle.battery.soc_max will be used as soc_target.
    """

    def __init__(
        self,
        env,
        ID,
        ismandatory=True,
        vehicle_filter=None,
        required_resources=None,
        resume=True,
        priority=0,
        recall_priority=-1,
        preempt=True,
        cancellable_for_dispatch=False,
        efficiency=1,
        soc_target=1,
    ):
        super(Charge, self).__init__(
            env,
            ID,
            ismandatory,
            vehicle_filter,
            required_resources,
            resume,
            priority,
            recall_priority,
            preempt,
            cancellable_for_dispatch,
            efficiency,
        )

        self.check_soc_target(soc_target, vehicle_filter)
        self.soc_target = soc_target

    @property
    def power(self):
        return self.charging_interface.max_power

    def _action(self, *args, **kwargs):
        self.vehicle.battery_logs.append(
            BatteryLog(self.env.now, self.vehicle, "charge_start")
        )
        if globalConstants["depot"]["log_cm_data"] and len(self.starts) == 1:
            # Log the first start only (ignore interruptions)
            self.vehicle.dwd.current_depot.evaluation.cm_report.log(
                ChargeStart(self.env, self.vehicle)
            )

        self.vehicle.battery.active_processes.append(self)

        if self.soc_target == "soc_max":
            self.soc_target = self.vehicle.battery.soc_max
        amount = (
            self.vehicle.battery.energy_real * self.soc_target
            - self.vehicle.battery.energy
        )
        assert amount > 0, (
            self.recall_count,
            self.vehicle.ID,
            self.vehicle.battery.energy,
            self.soc_target,
        )

        effective_power = self.power * self.efficiency
        self.dur = int(amount / effective_power * 3600)
        # flexprint('\t%s (energy=%d, energy_real=%d) needs to charge %f kWh. Will take %d s at %d kW'
        #           % (self.vehicle.ID, self.vehicle.battery.energy, self.vehicle.battery.energy_real, amount, self.dur, power),
        #           env=self.env, switch='operations')

        self.charging_interface.current_power = self.power
        self.vehicle.power_logs[self.env.now] = effective_power

        try:
            yield self.env.timeout(self.dur)

            # Update with exact amount to prevent rounding errors
            rest = round(
                self.vehicle.battery.energy_real * self.soc_target
                - self.vehicle.battery.energy,
                14,
            )
            assert rest >= 0
            if rest > 0:
                # rest is 0 if update_battery was called first in this same
                # second. Skip adding rest in this case because update_battery
                # already did
                self.update_battery("charge_end", amount=rest)

            # charge full event
            if globalConstants["depot"]["log_cm_data"]:
                self.vehicle.dwd.current_depot.evaluation.cm_report.log(
                    FullyCharged(self.env, self.vehicle)
                )

            flexprint(
                "\t%s completed charging." % (self.vehicle.ID),
                env=self.env,
                switch="operations",
            )

        except simpy.Interrupt:
            flexprint("charge interrupted", env=self.env, switch="processes")
            actual_charging_duration = self.env.now - self.starts[0]
            actual_charged_energy = (effective_power * actual_charging_duration) / 3600
            self.update_battery("charge_interrupt", amount=actual_charged_energy)

        self.charging_interface.current_power = 0
        self.vehicle.power_logs[self.env.now] = 0
        self.vehicle.battery.active_processes.remove(self)
        self.vehicle.battery.n_charges += 1

    @staticmethod
    def estimate_duration(vehicle, charging_interface, *args, **kwargs):
        """Return a duration estimate [int] assuming constant charging at.

        maximum power until soc_target is reached.
        """
        chargedata = Charge.get_chargedata(vehicle)
        if chargedata is None:
            # vehicle doesn't need charging (anymore) since vehicle filter
            # returned False
            return 0

        efficiency = chargedata["kwargs"]["efficiency"]
        soc_target = chargedata["kwargs"]["soc_target"]
        if soc_target == "soc_max":
            soc_target = vehicle.battery.soc_max

        amount = vehicle.battery.energy_real * soc_target - vehicle.battery.energy
        dur = int(amount / charging_interface.max_power * 3600 * efficiency)
        return dur

    @staticmethod
    def check_soc_target(soc_target, vehicle_filter):
        if soc_target != "soc_max" and not 0 < soc_target <= 1:
            raise ValueError("soc_target must be > 0 and <= 1 or 'soc_max'")
        if (
            "soc_lower_than" in vehicle_filter.filter_names
            and vehicle_filter.soc is not None
            and vehicle_filter.soc > soc_target
        ):
            raise ValueError(
                "soc_target for charging must be higher than "
                "the soc of filter_soc_lower_than"
            )


class ChargeSteps(ChargeAbstract):
    """Process of charging a vehicle's battery in linear power steps.

    Parameters:
    steps: [list] of lists or tuples that contain pairs of SoC and power.
        Example: [(0.8, 150), (1, 20)]
        The pairs define the power for charging a battery until the
        corresponding SoC value is reached. The above example states: "Charge
        at 150 kW until SoC=0.8, then reduce to 20 kW until the SoC=1".
        The process starts at the correct SoC value in steps, meaning that if
        in the example a vehicle that already has an SoC of 0.81 starts
        charging will start to charge with 20 kW.
        steps must contain at least one pair and must be sorted by SoC values
        in ascending order. Charging stops when the highest SoC value is
        reached, meaning that for steps=[(0.8, 50), (0.9, 150)] charging will
        stop at SoC=0.9. This functionality replaces the parameter 'soc_target'
        from the class ChargeAbstract/Charge.
        Note that the power values must not be higher than the maximum power of
        the charging interface.
    """

    def __init__(
        self,
        env,
        ID,
        steps,
        ismandatory=True,
        vehicle_filter=None,
        required_resources=None,
        resume=True,
        priority=0,
        recall_priority=-1,
        preempt=True,
        cancellable_for_dispatch=False,
        efficiency=1,
    ):
        super(ChargeSteps, self).__init__(
            env,
            ID,
            ismandatory,
            vehicle_filter,
            required_resources,
            resume,
            priority,
            recall_priority,
            preempt,
            cancellable_for_dispatch,
            efficiency,
        )

        self.steps = steps
        self._power = 0

    @property
    def power(self):
        return self._power

    @power.setter
    def power(self, value):
        self._power = value

    def _action(self, *args, **kwargs):
        self.vehicle.battery_logs.append(
            BatteryLog(self.env.now, self.vehicle, "charge_start")
        )
        if globalConstants["depot"]["log_cm_data"] and len(self.starts) == 1:
            # Log the first start only (ignore interruptions)
            self.vehicle.dwd.current_depot.evaluation.cm_report.log(
                ChargeStart(self.env, self.vehicle)
            )

        self.vehicle.battery.active_processes.append(self)

        # Find the entry point in steps
        steps_cropped = []
        for entry in self.steps:
            if entry[0] > self.vehicle.battery.soc:
                steps_cropped.append(entry)
        assert steps_cropped, (
            "soc is already higher than this process "
            "can achieve. Case should be avoided with "
            "a vehicle filter."
        )

        try:
            # Charge according to steps_cropped until steps_cropped is empty
            while steps_cropped:
                soc_step, self.power = steps_cropped.pop(0)
                amount = (
                    self.vehicle.battery.energy_real * soc_step
                    - self.vehicle.battery.energy
                )
                assert amount > 0, (
                    self.recall_count,
                    self.vehicle.ID,
                    self.vehicle.battery.energy,
                    soc_step,
                )
                effective_power = self.power * self.efficiency
                self.dur = int(amount / effective_power * 3600)

                # flexprint('\t%s (energy=%d, energy_real=%d) needs to charge %f kWh. Will take %d s at %d kW'
                #           % (self.vehicle.ID, self.vehicle.battery.energy, self.vehicle.battery.energy_real, amount, self.dur, power),
                #           env=self.env, switch='operations')

                self.charging_interface.current_power = self.power
                self.vehicle.power_logs[self.env.now] = effective_power
                yield self.env.timeout(self.dur)

                if steps_cropped:
                    self.update_battery("charge_step")
                    flexprint(
                        "Vehicle %s after update: soc_step=%s, soc: %s, energy=%s, energy_real=%s, amount=%s, power=%s, effective_power=%s, dur=%s, steps_cropped=%s, slot=%s"
                        % (
                            self.vehicle.ID,
                            soc_step,
                            self.vehicle.battery.soc,
                            self.vehicle.battery.energy,
                            self.vehicle.battery.energy_real,
                            amount,
                            self.power,
                            effective_power,
                            self.dur,
                            steps_cropped,
                            self.vehicle.dwd.current_slot,
                        ),
                        env=self.env,
                        switch="objID",
                        objID=self.vehicle.ID,
                    )

            rest = round(
                self.vehicle.battery.energy_real * soc_step
                - self.vehicle.battery.energy,
                14,
            )
            assert rest >= 0
            if rest > 0:
                # rest is 0 if update_battery was called first in this same
                # second. Skip adding rest in this case because update_battery
                # already did
                self.update_battery("charge_end", amount=rest)

            flexprint(
                "Vehicle %s after update: soc_step=%s, soc: %s, energy=%s, energy_real=%s, amount=%s, power=%s, effective_power=%s, dur=%s, steps_cropped=%s, slot=%s"
                % (
                    self.vehicle.ID,
                    soc_step,
                    self.vehicle.battery.soc,
                    self.vehicle.battery.energy,
                    self.vehicle.battery.energy_real,
                    amount,
                    self.power,
                    effective_power,
                    self.dur,
                    steps_cropped,
                    self.vehicle.dwd.current_slot,
                ),
                env=self.env,
                switch="objID",
                objID=self.vehicle.ID,
            )

            # charge full event
            if globalConstants["depot"]["log_cm_data"]:
                self.vehicle.dwd.current_depot.evaluation.cm_report.log(
                    FullyCharged(self.env, self.vehicle)
                )

            flexprint(
                "\t%s completed charging." % (self.vehicle.ID),
                env=self.env,
                switch="operations",
            )

        except simpy.Interrupt:
            flexprint("charge interrupted", env=self.env, switch="processes")
            self.update_battery("charge_interrupt")

        self.charging_interface.current_power = 0
        self.vehicle.power_logs[self.env.now] = 0
        self.vehicle.battery.active_processes.remove(self)
        self.vehicle.battery.n_charges += 1

    @staticmethod
    def estimate_duration(vehicle, *args, **kwargs):
        """Return a duration estimate [int] respecting steps."""
        chargedata = ChargeSteps.get_chargedata(vehicle)
        if chargedata is None:
            # vehicle doesn't need charging (anymore) since vehicle filter
            # returned False
            return 0

        efficiency = chargedata["kwargs"]["efficiency"]
        steps = chargedata["kwargs"]["steps"]

        steps_cropped = []
        for entry in steps:
            if entry[0] > vehicle.battery.soc:
                steps_cropped.append(entry)

        if not steps_cropped:
            return 0

        dur = 0
        while steps_cropped:
            soc_step, power = steps_cropped.pop(0)
            amount = vehicle.battery.energy_real * soc_step - vehicle.battery.energy
            effective_power = power * efficiency
            dur += int(amount / effective_power * 3600)

        return dur

    @staticmethod
    def check_steps(steps):
        """Check *steps* for validity."""
        try:
            iterable = iter(steps)
        except TypeError:
            raise TypeError("steps must be iterable.")
        else:
            if not steps:
                raise ValueError("steps cannot be empty.")
            previous = -1
            for soc, power in iterable:
                if previous >= soc:
                    raise ValueError(
                        "Entries in steps must be sorted by SoC "
                        "values in ascending order and cannot "
                        "contain equal SoC values."
                    )
                previous = soc


class ChargeEquationSteps(ChargeAbstract):
    """Process of charging a vehicle's battery based on a linearized function.

    for power.

    Parameters:
    peq_name: [str] name of equation that returns power based on the current
        situation. Must be in module eflips.depot.processes.
        Example: 'exponential_power'
    peq_params: {dict} of parameters that peq requires in addition to what is
        accessible via arguments.
    precision: [float] SoC invervals to call peq at while charging. Must
        be > 0 and < 1. Lower value means higher precision. High precision can
        significantly increase simulation runtime. Precision is capped to
        intervals resulting in step durations >= 1 second.
    soc_target: same as for class Charge
    """

    def __init__(
        self,
        env,
        ID,
        peq_name,
        peq_params,
        ismandatory=True,
        vehicle_filter=None,
        required_resources=None,
        resume=True,
        priority=0,
        recall_priority=-1,
        preempt=True,
        cancellable_for_dispatch=False,
        efficiency=1,
        precision=0.01,
        soc_target=1,
    ):
        super(ChargeEquationSteps, self).__init__(
            env,
            ID,
            ismandatory,
            vehicle_filter,
            required_resources,
            resume,
            priority,
            recall_priority,
            preempt,
            cancellable_for_dispatch,
            efficiency,
        )

        self.peq = getattr(eflips.depot.processes, peq_name)
        self.peq_params = peq_params
        if not 0 < precision < 1:
            raise ValueError("precision must be > 0 and < 1.")
        self.precision = precision
        Charge.check_soc_target(soc_target, vehicle_filter)
        self.soc_target = soc_target

    @property
    def power(self):
        return self.peq(self.vehicle, self.charging_interface, self.peq_params)

    def _action(self, *args, **kwargs):
        self.vehicle.battery_logs.append(
            BatteryLog(self.env.now, self.vehicle, "charge_start")
        )
        if globalConstants["depot"]["log_cm_data"] and len(self.starts) == 1:
            # Log the first start only (ignore interruptions)
            self.vehicle.dwd.current_depot.evaluation.cm_report.log(
                ChargeStart(self.env, self.vehicle)
            )

        self.vehicle.battery.active_processes.append(self)

        if self.soc_target == "soc_max":
            self.soc_target = self.vehicle.battery.soc_max
        assert (
            self.vehicle.battery.soc < self.soc_target
        ), "soc is already higher than this process can achieve. Case should be avoided with a vehicle filter."

        soc_target_step = 0
        try:
            while soc_target_step < self.soc_target:
                soc_interval = min(
                    self.precision, self.soc_target - self.vehicle.battery.soc
                )
                soc_target_step = self.vehicle.battery.soc + soc_interval
                amount = self.vehicle.battery.energy_real * soc_interval
                effective_power = self.power * self.efficiency
                self.dur = int(amount / effective_power * 3600)

                if self.dur == 0:
                    # amount is so small that dur is <1s. Reduce the precision
                    # to what 1 second would yield. Prevents yielding 0-
                    # timeouts with minimal amounts until the battery is full.
                    # flexprint('%s, soc: %s, soc_target: %s, soc_target_step: %s, soc_interval: %s, power: %s, dur: %s, amount: %s'
                    #       % (self.vehicle, self.vehicle.battery.soc, self.soc_target, soc_target_step, soc_interval, effective_power, self.dur, amount), env=self.env)
                    self.dur = 1
                    amount_1s = 1 / 3600 * effective_power
                    soc_interval_1s = amount_1s / self.vehicle.battery.energy_real
                    soc_target_1s = self.vehicle.battery.soc + soc_interval_1s
                    soc_target_step = min(soc_target_1s, self.soc_target)

                # flexprint('\t%s (energy=%d, energy_real=%d) needs to charge %f kWh. Will take %d s at %d kW'
                #           % (self.vehicle.ID, self.vehicle.battery.energy, self.vehicle.battery.energy_real, amount, self.dur, power),
                #           env=self.env, switch='operations')

                self.charging_interface.current_power = self.power
                self.vehicle.power_logs[self.env.now] = effective_power
                yield self.env.timeout(self.dur)

                if soc_target_step < self.soc_target:
                    # recalculate amount because update_battery may have been
                    # called in the meantime
                    amount_step = (
                        self.vehicle.battery.energy_real * soc_target_step
                        - self.vehicle.battery.energy
                    )
                    self.update_battery("charge_step", amount=amount_step)
                    flexprint(
                        "Vehicle %s after update: until_soc=%s, soc: %s, energy=%s, energy_real=%s, amount=%s, power=%s, effective_power=%s, dur=%s, slot=%s"
                        % (
                            self.vehicle.ID,
                            soc_interval,
                            self.vehicle.battery.soc,
                            self.vehicle.battery.energy,
                            self.vehicle.battery.energy_real,
                            amount,
                            self.power,
                            effective_power,
                            self.dur,
                            self.vehicle.dwd.current_slot,
                        ),
                        env=self.env,
                        switch="objID",
                        objID=self.vehicle.ID,
                    )
                else:
                    # Skip to adding rest
                    break

            rest = round(
                self.vehicle.battery.energy_real * self.soc_target
                - self.vehicle.battery.energy,
                14,
            )
            # assert rest >= 0, '%s, soc: %s, soc_target: %s, rest: %s, soc_target_step: %s, now: %s, last_update: %s, last log t: %s, last log name: %s' \
            #                   % (self.vehicle, self.vehicle.battery.soc, self.soc_target, rest, soc_target_step, self.env.now, self.last_update, self.vehicle.battery_logs[-1].t, self.vehicle.battery_logs[-1].event_name)
            if rest > 0:
                # rest is 0 if update_battery was called first in this same
                # second. Skip adding rest in this case because update_battery
                # already did
                self.update_battery("charge_end", amount=rest)

            # flexprint(
            #     'Vehicle %s after update: until_soc=%s, soc: %s, energy=%s, energy_real=%s, amount=%s, power=%s, effective_power=%s, dur=%s, slot=%s'
            #     % (self.vehicle.ID, soc_interval, self.vehicle.battery.soc, self.vehicle.battery.energy, self.vehicle.battery.energy_real, amount, self.power,
            #        effective_power, self.dur, self.vehicle.dwd.current_slot),
            #     env=self.env, switch='objID', objID=self.vehicle.ID)

            # charge full event
            if globalConstants["depot"]["log_cm_data"]:
                self.vehicle.dwd.current_depot.evaluation.cm_report.log(
                    FullyCharged(self.env, self.vehicle)
                )

            flexprint(
                "\t%s completed charging." % (self.vehicle.ID),
                env=self.env,
                switch="operations",
            )

        except simpy.Interrupt:
            flexprint("charge interrupted", env=self.env, switch="processes")
            self.update_battery("charge_interrupt")
            warnings.warn(
                "It is unclear whether interrupting a charging process of this type returns the desired "
                "result. Double-check the battery state after the simulation."
            )

        self.charging_interface.current_power = 0
        self.vehicle.power_logs[self.env.now] = 0
        self.vehicle.battery.active_processes.remove(self)
        self.vehicle.battery.n_charges += 1

    @staticmethod
    def estimate_duration(vehicle, charging_interface, *args, **kwargs):
        """Return a duration estimate [int] respecting steps."""
        chargedata = ChargeEquationSteps.get_chargedata(vehicle)
        if chargedata is None:
            # vehicle doesn't need charging (anymore) since vehicle filter
            # returned False
            return 0

        efficiency = chargedata["kwargs"]["efficiency"]
        soc_target = chargedata["kwargs"]["soc_target"]
        precision = chargedata["kwargs"]["precision"]
        peq = getattr(eflips.depot.processes, chargedata["kwargs"]["peq_name"])
        peq_params = chargedata["kwargs"]["peq_params"]
        if soc_target == "soc_max":
            soc_target = vehicle.battery.soc_max

        soc = vehicle.battery.soc
        dur = 0

        while soc < soc_target:
            soc_interval = min(precision, soc_target - soc)
            soc += soc_interval

            amount = vehicle.battery.energy * soc_interval

            power = peq(vehicle, charging_interface, peq_params)
            effective_power = power * efficiency
            dur += amount / effective_power * 3600

        return int(dur)


def exponential_power(vehicle, charging_interface, peq_params, *args, **kwargs):
    """Return power in kW for the use with ChargeEquationSteps.

    Parameters:
    vehicle: [SimpleVehicle]
    charging_interface: [DepotChargingInterface]
    peq_params: [dict] with parameters for this function in addition to what is
        accessible via vehicle and charging_interface
    """
    SoC = vehicle.battery.soc
    P_max = charging_interface.max_power
    SoC_threshold = peq_params["x"]
    En = vehicle.battery.energy_nominal

    if SoC < SoC_threshold:
        return P_max
    else:
        # return P_max ** (5 - SoC / 0.2) * 10 ** (4 - SoC / 0.2) * En ** (SoC / 0.2 - 4)
        return (
            P_max
            / 10
            * (1 / En - 10)
            / (math.exp(1) - math.exp(SoC_threshold))
            * (math.exp(SoC) - math.exp(SoC_threshold))
            + P_max
        )


class Standby(VehicleProcess):
    """Process of mandatory waiting such as standby times."""

    def __init__(
        self,
        env,
        ID,
        dur,
        ismandatory=True,
        vehicle_filter=None,
        required_resources=None,
        resume=True,
        priority=0,
        recall_priority=-1,
        preempt=True,
        cancellable_for_dispatch=False,
    ):
        super(Standby, self).__init__(
            env,
            ID,
            dur,
            ismandatory,
            vehicle_filter,
            required_resources,
            resume,
            priority,
            recall_priority,
            preempt,
            cancellable_for_dispatch,
        )

    def _action(self, vehicle):
        try:
            flexprint(
                "\t%s started standby." % (vehicle.ID),
                env=self.env,
                switch="operations",
            )
            yield self.env.timeout(self.dur)
            flexprint(
                "\t%s completed standby." % (vehicle.ID),
                env=self.env,
                switch="operations",
            )

        except simpy.Interrupt:
            flexprint("action interrupted", env=self.env, switch="processes")
            pass

    @staticmethod
    def estimate_duration(dur, *args, **kwargs):
        return dur


class Repair(VehicleProcess):
    """Process of repairing a vehicle."""

    def __init__(
        self,
        env,
        ID,
        dur,
        ismandatory=True,
        vehicle_filter=None,
        required_resources=None,
        resume=True,
        priority=0,
        recall_priority=-1,
        preempt=True,
        cancellable_for_dispatch=False,
    ):
        super(Repair, self).__init__(
            env,
            ID,
            dur,
            ismandatory,
            vehicle_filter,
            required_resources,
            resume,
            priority,
            recall_priority,
            preempt,
            cancellable_for_dispatch,
        )

    def _action(self, vehicle):
        try:
            flexprint(
                "\t%s started repair." % (vehicle.ID), env=self.env, switch="operations"
            )
            yield self.env.timeout(self.dur)
            vehicle.dwd.repair_need = False
            flexprint(
                "\t%s completed repair." % (vehicle.ID),
                env=self.env,
                switch="operations",
            )

        except simpy.Interrupt:
            flexprint("action interrupted", env=self.env, switch="processes")
            pass

    @staticmethod
    def estimate_duration(dur, *args, **kwargs):
        return dur


class Maintain(VehicleProcess):
    """Process of maintaining a vehicle."""

    def __init__(
        self,
        env,
        ID,
        dur,
        ismandatory=True,
        vehicle_filter=None,
        required_resources=None,
        resume=True,
        priority=0,
        recall_priority=-1,
        preempt=True,
        cancellable_for_dispatch=False,
    ):
        super(Maintain, self).__init__(
            env,
            ID,
            dur,
            ismandatory,
            vehicle_filter,
            required_resources,
            resume,
            priority,
            recall_priority,
            preempt,
            cancellable_for_dispatch,
        )

    def _action(self, vehicle):
        try:
            flexprint(
                "\t%s started maintenance." % (vehicle.ID),
                env=self.env,
                switch="operations",
            )
            yield self.env.timeout(self.dur)
            vehicle.dwd.maintenance_need = False
            flexprint(
                "\t%s completed maintenance." % (vehicle.ID),
                env=self.env,
                switch="operations",
            )

        except simpy.Interrupt:
            flexprint("action interrupted", env=self.env, switch="processes")
            pass

    @staticmethod
    def estimate_duration(dur, *args, **kwargs):
        return dur


class Precondition(VehicleProcess):
    """Preconditioning for a vehicle.

    Parameters:
    required_resources: doesn't need to contain the charging interface because
        it's detected automatically.
    power: in kW.

    Attributes:
    request_immediately: is False for this class because preconditioning is
        scheduled depending on departure time instead of entry time at an area.
    """

    request_immediately = False

    def __init__(
        self,
        env,
        ID,
        dur,
        ismandatory=False,
        vehicle_filter=None,
        required_resources=None,
        resume=True,
        priority=0,
        recall_priority=-1,
        preempt=True,
        cancellable_for_dispatch=False,
        efficiency=1,
        power=20,
    ):
        super(Precondition, self).__init__(
            env,
            ID,
            dur,
            ismandatory,
            vehicle_filter,
            required_resources,
            resume,
            priority,
            recall_priority,
            preempt,
            cancellable_for_dispatch,
        )

        self.efficiency = efficiency
        self.power = power

    def _action(self, vehicle):
        try:
            flexprint(
                "\t%s started preconditioning." % (vehicle.ID),
                env=self.env,
                switch="operations",
            )
            effective_power = self.power * self.efficiency
            self.charging_interface.current_power = self.power
            self.vehicle.power_logs[self.env.now] = effective_power
            yield self.env.timeout(self.dur)
            self.charging_interface.current_power = 0
            self.vehicle.power_logs[self.env.now] = 0

            flexprint(
                "\t%s completed preconditioning." % (vehicle.ID),
                env=self.env,
                switch="operations",
            )

        except simpy.Interrupt:
            flexprint("action interrupted", env=self.env, switch="processes")
            self.charging_interface.current_power = 0
            self.vehicle.power_logs[self.env.now] = 0

    def _pre(self, *args, **kwargs):
        super()._pre(*args, **kwargs)
        self.charging_interface = self.vehicle.dwd.current_charging_interface

    @staticmethod
    def estimate_duration(dur, *args, **kwargs):
        return dur



# -*- coding: utf-8 -*-
"""
Example plot calls and evaluation after(!) simulation.
For plot details see evaluation.py.
"""

# Plots and results for evaluation of depot layout

ev.vehicle_periods(
    periods={
        "depot general": "darkgray",
        "park": "lightgray",
        "serve_supply_clean_daily": "steelblue",
        "serve_clean_ext": "darkblue",
        "charge_dc": "forestgreen",
        "charge_oc": "forestgreen",
        "precondition": "black",
    },
    save=True,
    show=True,
    formats=(
        "pdf",
        "png",
    ),
    show_total_power=True,
    show_annotates=True,
)

ev.nvehicles_used(save=True, show=True)
ev.departure_delay_vt(save=True, show=True)

ev.sl_all(save=True, show=True)

ev.nvehicles_initstore_end(save=True, show=True)
ev.nvehicles_area("S_1", save=True, show=True)

ev.usage_history("workers_service", save=True, show=True)
ev.usage_history("shunting_service", save=True, show=True)

ev.area_history("SB_DC 1", save=True, show=True)
ev.battery_level("SB_DC 1", save=True, show=True)
ev.vehicle_power("SB_DC 1", save=True, show=True)

ev.idle_time_dist(charge_IDs=("charge_dc", "charge_oc"), bins=60, save=True, show=True)

ev.arrival_soc(
    typs={"SB_DC": "#2ca02c", "AB_OC": "#CD3333"},
    formats=(
        "pdf",
        "png",
    ),
    save=True,
    show=True,
    language="de",
)
ev.departure_soc(
    typs={"SB_DC": "#2ca02c", "AB_OC": "#CD3333"},
    formats=(
        "pdf",
        "png",
    ),
    save=True,
    show=True,
    language="de",
)

ev.congestion(save=True, show=True)
ev.occupancy_rate()

ev.lead_time_match_scatter(save=True, show=True)
ev.lead_time_match_dist(save=True, show=True)

for resource in ev.depot.resources:
    if resource[0:2] == "ci":
        ci_no_str = resource[3:]
        ev.ci_power(cis=[int(ci_no_str), int(ci_no_str)])

for ci_no in range(1, 20):
    ev.ci_power(cis=[ci_no, ci_no])


# Plots and results for evaluation of depot layout with smart charging
ev.vehicle_periods(
    periods={
        "depot general": "darkgray",
        "park": "lightgray",
        "serve_supply": "skyblue",
        "serve_clean_daily": "steelblue",
        "serve_clean_ext": "darkblue",
        "precondition": "black",
    },
    basefilename="vehicle_periods_sc",
    save=True,
    show=False,
    formats=(
        "pdf",
        "png",
    ),
    show_total_power=True,
    smart_charging=smart_charging,
    smart_charging_color="forestgreen",
)

smart_charging.results()
smart_charging.plot_results()



"""Utilities for converting config data, config names, templates and visu."""
import json
from eflips.depot.layout_opt import packing


def save_json(obj, filename):
    """Write python object *obj* to a json file. Overwrite file if it already
    exists (without comfirmation prompt).
    filename: [str] excluding file extension
    """
    filename = filename + ".json"
    with open(filename, "w") as file:
        json.dump(obj, file, indent=4)


def name_from_config(area_config):
    """Create a name [str] from area_config [dict]."""
    name = ""

    for typename_short, data in area_config.items():
        for capacity, n in data.items():
            name += "_" + str(n) + "x" + str(capacity) + typename_short

    if name:
        name = name[1:]

    return name


def config_from_name(name):
    """Create an area config [dict] from name [str]."""

    def decode(a):
        n = int(a[: a.index("x")])
        c = int(a[a.index("x") + 1 : -1])
        t = a[-1]
        return n, c, t

    ac = {"d": {}, "l": {}}
    for ai in name.split("_"):
        ni, ci, ti = decode(ai)
        ac[ti][ci] = ni

    return ac


def config_from_name_3a(name):
    """Create an area config [dict] from name [str] where types are DDR, DSR, DSR_90
    L.
    """

    def decode(a):
        i_lastdigit = a.rfind(next(si for si in a[::-1] if si.isdigit()))
        n = int(a[: a.index("x")])
        c = int(a[a.index("x") + 1 : i_lastdigit + 1])
        t = a[i_lastdigit + 1 :]
        return n, c, t

    ac = {"DDR": {}, "DSR": {}, "DSR_90": {}, "L": {}}
    for ai in name.split("_"):
        ni, ci, ti = decode(ai)
        ac[ti][ci] = ni

    return ac


def template_from_config(area_config):
    parking_capacity = 0
    for data in area_config.values():
        for capacity, n in data.items():
            for ni in range(n):
                parking_capacity += capacity

    ID_counters = {"ci": 0, "pa": 0}

    template = {
        "templatename_display": "",
        "general": {"depotID": "", "dispatch_strategy_name": ""},
        "resources": {},
        "resource_switches": {},
        "processes": {},
        "areas": {},
        "groups": {},
        "plans": {},
    }

    template["general"]["depotID"] = "KLS"
    template["general"]["dispatch_strategy_name"] = "SMART"

    template["resources"]["workers_service"] = {
        "typename": "DepotResource",
        "capacity": capacity_service,
    }

    template["resource_switches"]["service_switch"] = {
        "resource": "workers_service",
        "breaks": [[25200, 61200]],
        "preempt": True,
        "strength": "full",
    }

    template["processes"]["charge"] = {
        "typename": "Charge",
        "ismandatory": False,
        "vehicle_filter": {"filter_names": ["soc_lower_than"], "soc": 0.9},
        "cancellable_for_dispatch": False,
        "soc_target": 0.9,
        "efficiency": 0.95,
    }

    template["processes"]["serve"] = {
        "typename": "Serve",
        "dur": 600,
        "ismandatory": False,
        "vehicle_filter": {"filter_names": ["in_period"], "period": [57600, 20700]},
        "required_resources": ["workers_service"],
        "cancellable_for_dispatch": False,
    }

    template["processes"]["standby_arr"] = {
        "typename": "Standby",
        "dur": 300,
        "ismandatory": True,
        "vehicle_filter": None,
        "required_resources": [],
        "cancellable_for_dispatch": False,
    }

    template["processes"]["standby_dep"] = {
        "typename": "Standby",
        "dur": 900,
        "ismandatory": True,
        "vehicle_filter": None,
        "required_resources": [],
        "cancellable_for_dispatch": False,
    }

    # Areas
    template["areas"]["Stauflaeche"] = {
        "typename": "DirectArea",
        "capacity": parking_capacity,
        "available_processes": ["standby_arr"],
        "issink": False,
        "entry_filter": None,
    }

    template["areas"]["Serviceflaeche"] = {
        "typename": "DirectArea",
        "capacity": capacity_service,
        "available_processes": ["serve"],
        "issink": False,
        "entry_filter": None,
    }

    # Parking areas
    def add_parking_area_direct(capacity):
        ID_counters["pa"] += 1
        ID = "area_" + str(ID_counters["pa"])
        cis = []
        for ci in range(capacity):
            cis.append("ci_" + str(ID_counters["ci"]))
            ID_counters["ci"] += 1

        template["areas"][ID] = {
            "typename": "DirectArea",
            "amount": 1,
            "capacity": capacity,
            "charging_interfaces": cis,
            "available_processes": ["charge", "standby_dep"],
            "issink": True,
            "entry_filter": None,
        }

    def add_parking_area_line(capacity):
        ID_counters["pa"] += 1
        ID = "area_" + str(ID_counters["pa"])
        cis = []
        for ci in range(capacity):
            cis.append("ci_" + str(ID_counters["ci"]))
            ID_counters["ci"] += 1

        template["areas"][ID] = {
            "typename": "LineArea",
            "amount": 1,
            "capacity": capacity,
            "charging_interfaces": cis,
            "available_processes": ["charge", "standby_dep"],
            "issink": True,
            "entry_filter": None,
            "side_put_default": "front",
            "side_get_default": "back",
        }

    for typename_short, data in area_config.items():
        for capacity, n in data.items():
            for ni in range(n):
                if typename_short == "d":
                    add_parking_area_direct(capacity)
                else:
                    add_parking_area_line(capacity)

    # Charging interfaces
    for i in range(ID_counters["ci"]):
        ID = "ci_" + str(i)
        template["resources"][ID] = {
            "typename": "DepotChargingInterface",
            "max_power": power,
        }

    # Charging switches
    for i in range(ID_counters["ci"]):
        ID = "cs_" + str(i)
        template["resource_switches"][ID] = {
            "resource": "ci_" + str(i),
            "breaks": [[64800, 72000]],
            "preempt": True,
            "strength": "full",
        }

    # Groups
    template["groups"]["parking area group"] = {
        "typename": "ParkingAreaGroup",
        "stores": [key for key, value in template["areas"].items() if value["issink"]],
        "parking_strategy_name": "SMART2",
    }

    # Plan
    template["plans"]["default"] = {
        "typename": "DefaultActivityPlan",
        "locations": ["Stauflaeche", "Serviceflaeche", "parking area group"],
    }

    name = name_from_config(area_config)
    template["templatename_display"] = name

    save_json(template, "templates\\" + name)
    print("Saved ", name)


def visu_from_config(area_config, a, b):
    """Only works for config with visu keys listed below."""
    visus = {
        "DDR": packing.VisuDataDirectDoubleRow,
        "DSR": packing.VisuDataDirectSingleRow,
        "DSR_90": packing.VisuDataDirectSingleRow_90,
        "L": packing.VisuDataLine,
    }

    bin = packing.BinWithDistances(a, b)
    for typename, data in area_config.items():
        for c, n in data.items():
            for i in range(n):
                bin.items.append(visus[typename](capacity=c))

    return bin


def ind_from_config(area_config, creator):
    """Create an individual [DepotPrototype] from area_config [dict].
    Requires a the deap creator to be setup. Depot a and b are taken from
    depot_layout_opt.settings.OPT_CONSTANTS (needs to be loaded before calling
    this function)."""
    from eflips.depot.layout_opt.opt_tools.init import (
        DSRPrototype,
        DSR_90Prototype,
        DDRPrototype,
        LinePrototype,
    )

    protos = {
        "DDR": DDRPrototype,
        "DSR": DSRPrototype,
        "DSR_90": DSR_90Prototype,
        "L": LinePrototype,
    }
    ind = creator.Individual()
    for typename, data in area_config.items():
        for c, n in data.items():
            for i in range(n):
                ind.areas.append(protos[typename](capacity=c))
    return ind


def template_from_name(name):
    template_from_config(config_from_name(name))


if __name__ == "__main__":
    power = 150
    capacity_service = 0
    # area_config = {
    #     'd': {
    #         148: 1
    #     },
    #     'l': {
    #
    #     }
    # }

    # template_from_config(area_config, power=150, capacity_service=3)

    # template_from_name('2x10d_2x10l_1x4l')

    # area_config = {
    #     'DDR': {
    #
    #     },
    #     'DSR': {
    #         23: 1
    #     },
    #     'L': {
    #         8: 17
    #     }
    # }

    # import matplotlib.pyplot as plt
    # plt.rcParams.update({
    #         "font.family": "serif",
    #         "font.sans-serif": "Liberation Sans, Arial, Helvetica, DejaVu Sans",
    #         "font.serif": "Times new Roman",
    #         "axes.labelsize": 12,
    #         # "axes.titleweight": "normal",
    #         # "axes.titlesize": 12,
    #         # "legend.fontsize": 12,
    #         # "legend.frameon": True,
    #         # "figure.autolayout": True,
    #         # "figure.dpi": 300
    #     })
    # bin = visu_from_config(area_config, 88, 140)

    area_config = {"DDR": {65: 0}, "DSR": {33: 0}, "DSR_90": {33: 0}, "L": {4: 1}}
    # bin = visu_from_config(area_config, 47, 202)
    # bin.draw()

    from eflips.depot.layout_opt import optimize_c_urfd
    from eflips.depot.layout_opt.opt_tools import fitness_c_urfd

    ind = ind_from_config(area_config, optimize_c_urfd.creator)
    results = fitness_c_urfd.evaluate(ind)
    ind.fitness.values = results[:-1]
    ind.results = results[-1]
    print(fitness_c_urfd.feasible(ind))



from datetime import timedelta, datetime
from typing import List, Dict

import numpy as np
import scipy
from eflips.model import Event, EventType


def optimize_charging_events_even(charging_events: List[Event]) -> None:
    """
    This function optimizes the power draw of a list of charging events.

    The power draw is optimized such that the total
    power draw is minimized, while the energy transferred remains constant.
    :param charging_events: The list of charging events to optimize
    :return: Nothing, the charging events are updated in place
    """

    TEMPORAL_RESOLUTION = timedelta(seconds=60)

    assert all(
        [event.event_type == EventType.CHARGING_DEPOT for event in charging_events]
    )
    start_time = min([event.time_start for event in charging_events])
    end_time = max([event.time_end for event in charging_events])

    # Formulate the optimzation problem.
    # - Each charging event has a peak power, which cannot be exceeded.
    # - Each charging event has a start and end time.
    # - betweeen the start time and end time, the energy transferred must remain constant.
    # - the total power draw is the sum of all events' power at this point in time
    # - the total power draw must be minimized

    total_duration = int((end_time - start_time) / TEMPORAL_RESOLUTION)
    total_time = np.arange(
        start_time.timestamp(),
        end_time.timestamp(),
        TEMPORAL_RESOLUTION.total_seconds(),
    )  # The time axis, used to resample the power draw

    # For each event, create an array of power draws and a boolean array of charging allowed
    # Also note down the peak power and transferred energy
    params_for_events: List[Dict[str, float | np.ndarray | Event]] = []
    for event in charging_events:
        power_draw = np.zeros_like(total_time, dtype=float)
        charging_allowed = np.zeros_like(total_time, dtype=int)

        # Calculate the power draw vector, from the start SoC, end SoC and timeseries, if available
        event_soc = [event.soc_start]
        event_time = [event.time_start.timestamp()]
        if event.timeseries is not None:
            event_soc.extend(event.timeseries["soc"])
            event_time.extend(
                [
                    datetime.fromisoformat(t).timestamp()
                    for t in event.timeseries["time"]
                ]
            )
        event_soc.append(event.soc_end)
        event_time.append(event.time_end.timestamp())

        # Resample the timeseries to the temporal resolution
        expanded_soc = np.interp(total_time, event_time, event_soc)
        expanded_power = (
            np.diff(expanded_soc, prepend=expanded_soc[0])
            * event.vehicle.vehicle_type.battery_capacity
        )  # Change in kWh each minute
        expanded_power = (
            expanded_power / TEMPORAL_RESOLUTION.total_seconds() * 3600
        )  # to kW

        # Between the start and end time, charging is allowed
        start_index = int((event.time_start - start_time) / TEMPORAL_RESOLUTION)
        end_index = int((event.time_end - start_time) / TEMPORAL_RESOLUTION)
        charging_allowed[start_index:end_index] = 1
        max_power = max(expanded_power)
        transferred_energy = (
            event_soc[-1] - event_soc[0]
        ) * event.vehicle.vehicle_type.battery_capacity  # kWh

        params_for_events.append(
            {
                "event": event,
                "power_draw": expanded_power,
                "charging_allowed": charging_allowed,
                "max_power": max_power,
                "transferred_energy": transferred_energy,
            }
        )

    # The total number of vehicles at the depot is the sum of the charging allowed for each event, since it is 1 if
    # the bus is at the depot and 0 otherwise
    total_occupancy = np.sum(
        [event["charging_allowed"] for event in params_for_events], axis=0
    )

    # For each event, calculate an optimized power draw
    for params_for_event in params_for_events:
        # Calculate the mean power that would be drawn if the vehicle was charged evenly
        charging_duration = (
            np.sum(params_for_event["charging_allowed"])
            * TEMPORAL_RESOLUTION.total_seconds()
        )
        mean_power = (
            params_for_event["transferred_energy"] / charging_duration
        ) * 3600  # kW

        # Calculate the mean amount of vehicles present at the depot when this vehicle is charging
        # We only consider the time when the vehicle is charging
        total_occupancy_while_charging = total_occupancy[
            params_for_event["charging_allowed"] == 1
        ]
        mean_occupancy = np.mean(total_occupancy_while_charging)

        # for each timestep optimize the power draw. We want to charge less when there are more vehicles at the depot
        # The power draw is varies with the amount of vehicles present at the depot relative to the mean amount
        charging_factor = (total_occupancy / mean_occupancy) * params_for_event[
            "charging_allowed"
        ]

        # Make sure the charging factor is not infinite or NaN
        charging_factor[np.isnan(charging_factor)] = 1
        charging_factor[np.isinf(charging_factor)] = 1
        power_scaling_vector = (
            (mean_power * charging_factor) - mean_power
        ) * -1  # How much to shift the power draw
        if min(power_scaling_vector) < -mean_power:
            power_scaling_vector /= min(power_scaling_vector) / -mean_power
        optimized_power = params_for_event["charging_allowed"] * (
            power_scaling_vector + mean_power
        )

        # Cap it at the peak power
        optimized_power_capped = np.minimum(
            optimized_power, params_for_event["max_power"]
        )

        # Count the energy that is we need to distribute over the time when the vehicle is not at peak power
        energy_to_distribute = (
            np.trapz((optimized_power - optimized_power_capped), total_time) / 3600
        )  # kWh
        if energy_to_distribute > 0:
            optimized_power_not_capped = np.where(
                optimized_power > optimized_power_capped
            )
            # Distribute the energy over the time when the vehicle is not at peak power
            uncapped_duration = (
                optimized_power_not_capped[0].shape[0]
                * TEMPORAL_RESOLUTION.total_seconds()
            )
            power_to_add = energy_to_distribute / (uncapped_duration / 3600)
            optimized_power[optimized_power_not_capped[0]] += power_to_add

        # Make sure the transferred energy is the same
        post_opt_energy = (
            scipy.integrate.trapezoid(optimized_power, total_time) / 3600
        )  # kWh

        if False:
            # Some plots for only this charging event
            from matplotlib import pyplot as plt

            valid_charging_indices = np.where(
                params_for_event["charging_allowed"] == 1
            )[0]

            fig, axs = plt.subplots(3, 1, sharex=True)
            axs[0].axhline(mean_power, color="red", linestyle="--", label="Mean power")
            axs[0].plot(
                total_time[valid_charging_indices],
                params_for_event["power_draw"][valid_charging_indices],
                label="Original power draw",
            )
            axs[0].plot(
                total_time[valid_charging_indices],
                optimized_power[valid_charging_indices],
                label="Optimized power draw",
            )
            axs[0].plot(
                total_time[valid_charging_indices],
                optimized_power_capped[valid_charging_indices],
                label="Optimized power draw capped",
            )
            axs[0].set_xlabel("Time")
            axs[0].legend()

            axs[1].plot(
                total_time[valid_charging_indices],
                np.cumsum(params_for_event["power_draw"][valid_charging_indices])
                / 3600,
                label="Original energy transferred",
            )
            axs[1].plot(
                total_time[valid_charging_indices],
                np.cumsum(optimized_power[valid_charging_indices]) / 3600,
                label="Optimized energy transferred",
            )
            axs[1].set_xlabel("Time")

            axs[2].axhline(
                mean_occupancy, color="red", linestyle="--", label="Mean occupancy"
            )
            axs[2].plot(
                total_time[valid_charging_indices],
                total_occupancy[valid_charging_indices],
            )

            mean_arr = np.ones_like(total_occupancy) * mean_occupancy
            mean_cumsum = np.cumsum(mean_arr[valid_charging_indices])
            total_cumsum = np.cumsum(total_occupancy[valid_charging_indices])
            assert np.isclose(mean_cumsum[-1], total_cumsum[-1], atol=0.01)

            axs[2].set_xlabel("Time")
            plt.show()

        if not np.isclose(
            post_opt_energy, params_for_event["transferred_energy"], rtol=0.001
        ):
            # Scale the power draw to match the transferred energy
            optimized_power = optimized_power * (
                params_for_event["transferred_energy"] / post_opt_energy
            )

        optimized_power2 = (
            params_for_event["charging_allowed"] * mean_power
        )  # * (mean_occupancy / total_occupancy)
        # Fill NaNs with the zero power draw
        optimized_power[np.isnan(optimized_power)] = 0
        optimized_power2[np.isnan(optimized_power2)] = 0

        params_for_event["optimized_power"] = optimized_power
        params_for_event["optimized_power2"] = optimized_power2

        event = params_for_event["event"]
        start_index = int((event.time_start - start_time) / TEMPORAL_RESOLUTION)
        end_index = (
            int((event.time_end - start_time) / TEMPORAL_RESOLUTION) + 1
        )  # +1 to include the last index
        powers = params_for_event["optimized_power"][start_index:end_index]

        energies = scipy.integrate.cumulative_trapezoid(powers, initial=0) / (
            3600 / TEMPORAL_RESOLUTION.total_seconds()
        )  # kWh
        socs = event.soc_start + energies / event.vehicle.vehicle_type.battery_capacity

        # Make sure the last SoC is the same as the end SoC
        assert np.isclose(socs[-1], event.soc_end, atol=0.01)
        # Make sure the first SoC is the same as the start SoC
        assert np.isclose(socs[0], event.soc_start, atol=0.01)

        # Make the socs match exactly, setting all those smaller than the start SoC to the start SoC and
        # all those larger than the end SoC to the end SoC
        socs[socs < event.soc_start] = event.soc_start
        socs[socs > event.soc_end] = event.soc_end

        # Add a timeseries to the event, removing the first and last index, since they will be the same as the start and
        # end SoC
        event.timeseries = {
            "time": [
                datetime.fromtimestamp(t).astimezone().isoformat()
                for t in total_time[start_index:end_index]
            ][
                1:-1
            ],  # Remove the first and last index
            "soc": socs.tolist()[1:-1],
        }
        if event.timeseries["time"][0] < event.time_start.isoformat():
            event.timeseries["time"][0] = event.time_start.isoformat()
        if event.timeseries["time"][-1] > event.time_end.isoformat():
            event.timeseries["time"][-1] = event.time_end.isoformat()

    # Now we have the power draw and charging allowed for each event
    if False:
        from matplotlib import pyplot as plt

        fig, axs = plt.subplots(3, 1, sharex=True)
        total_power = np.sum(
            [event["power_draw"] for event in params_for_events], axis=0
        )
        axs[0].plot(total_time, total_power, label="Original power draw")
        optimized_power = np.sum(
            [event["optimized_power"] for event in params_for_events], axis=0
        )
        axs[0].plot(total_time, optimized_power, label="Optimized power draw")
        optimized_power2 = np.sum(
            [event["optimized_power2"] for event in params_for_events], axis=0
        )
        axs[0].plot(total_time, optimized_power2, label="Mean power draw")
        axs[0].set_xlabel("Time")
        axs[0].set_ylabel("Total power draw (kW)")

        axs[0].axhline(
            y=max(total_power), color="blue", linestyle="--", label="Max power draw"
        )
        axs[0].axhline(
            y=max(optimized_power),
            color="orange",
            linestyle="--",
            label="Max optimized power draw",
        )
        axs[0].axhline(
            y=max(optimized_power2),
            color="green",
            linestyle="--",
            label="Max mean power draw",
        )

        # Energy transferred
        total_energy = scipy.integrate.cumulative_trapezoid(
            total_power, total_time, initial=0
        )
        axs[1].plot(total_time, total_energy, label="Original energy transferred")
        optimized_energy = scipy.integrate.cumulative_trapezoid(
            optimized_power, total_time, initial=0
        )
        axs[1].plot(total_time, optimized_energy, label="Optimized energy transferred")
        optimized_energy2 = scipy.integrate.cumulative_trapezoid(
            optimized_power2, total_time, initial=0
        )
        axs[1].plot(total_time, optimized_energy2, label="Mean energy transferred")
        axs[1].set_xlabel("Time")
        axs[1].set_ylabel("Total energy transferred (kWh)")

        axs[2].plot(total_time, total_occupancy)
        axs[2].set_xlabel("Time")
        axs[2].set_ylabel("Vehicle count")
        plt.show()



from eflips.depot.layout_opt.opt_tools.crossover import (
    cxOnePoint_depot,
    cxTwoPoint_depot,
    cxMessyOnePoint_depot,
)
from eflips.depot.layout_opt.opt_tools.mutation import (
    mutgaussian_depot_and,
    mutgaussian_depot_or_partial,
    mutuniform_depot_or_partial,
    mutgaussian_depot_or,
    mutuniform_depot_or,
)
from eflips.depot.layout_opt.opt_tools.init import (
    DepotPrototype,
    clone_depot,
    init_random_depot,
    DSRPrototype,
    DSR_90Prototype,
    DDRPrototype,
    LinePrototype,
)
from eflips.depot.layout_opt.opt_tools.fitness_util import simulate



"""Methods to check and modify eflips settings.

Separate from eflips.settings.py as preparation for making settings
modifications possible through a GUI, similar as with
eflips.depot.configuration.py.
"""
from eflips.depot.simple_vehicle import VehicleType, VehicleTypeGroup
from eflips.settings import globalConstants as gc


def check_gc_validity():
    """Checks of globalConstants that raise an error if not passed."""

    # substitutable_types
    st = gc["depot"]["substitutable_types"]  # create shortcut
    if not isinstance(st, list) or not all(isinstance(group, list) for group in st):
        raise ValueError("Invalid globalConstants substitutable_types.")


def complete_gc():
    """Actions that must take place before simulation start, but may not
    be possible during the configuration phase.

    """
    # Create vehicle types and save in list and dict
    type_data = gc["depot"]["vehicle_types"]
    gc["depot"]["vehicle_types_obj"] = []
    vto = gc["depot"]["vehicle_types_obj"]
    gc["depot"]["vehicle_types_obj_dict"] = {}
    vto_dict = gc["depot"]["vehicle_types_obj_dict"]
    for ID in type_data:
        vt = VehicleType(ID, **type_data[ID])
        vto.append(vt)
        vto_dict[ID] = vt

    # Check vehicle type groups
    st = gc["depot"]["substitutable_types"]  # shortcut
    control = []
    for group_data in st:
        control.extend(group_data)
    if any(control.count(ID) > 1 for ID in control):
        raise ValueError(
            "Vehicle types cannot have multiple appearances in "
            "globalConstants['depot']['substitutable_types']."
        )

    # Create vehicle type groups
    gc["depot"]["vehicle_type_groups"] = []
    for group_data in st:
        types = [next(vt for vt in vto if vt.ID == ID) for ID in group_data]
        group = VehicleTypeGroup(types)
        gc["depot"]["vehicle_type_groups"].append(group)
        for vehicle_type in types:
            vehicle_type.group = group

    # Create a list of str representations of vehicle types and groups
    categories = []
    for vt in gc["depot"]["vehicle_types_obj"]:
        if vt.group is None:
            categories.append(vt.ID)
        else:
            categories.append(", ".join(vt.ID for vt in vt.group.types))
    categories = list(set(categories))
    categories.sort()
    gc["depot"]["vehicle_type_categories"] = categories

    # Precompute maximum battery capacity and save it in globalConstants
    gc["depot"]["max_battery_capacity"] = max(vt.battery_capacity for vt in vto)



"""
Validation tests for after executing the depot simulation.

"""
from collections import Counter


class Validator:
    """

    ev: [DepotEvaluation]
    """

    def __init__(self, ev):
        self.ev = ev
        self.results = {}

    @property
    def valid(self):
        return all(ri["valid"] for ri in self.results.values())

    def all_periods(self, periods_map):
        """Prepare and call all periods-related tests.

        periods_map: [dict] of period key and their template-specific name as
        str. For the charging process, the value is a list.
            E.g.:
            {'depot general': 'depot general',
            'park': 'park',
            'serve': 'serve',
            'charge': ['charge_dc', 'charge_oc']}
        """
        # Unpack names for getting data
        period_names = []
        for v in periods_map.values():
            if isinstance(v, str):
                period_names.append(v)
            else:
                for vi in v:
                    period_names.append(vi)

        # Get data
        vehicledata = {}
        vehicle_no = 0
        for vehicle in self.ev.vehicle_generator.items:
            vehicledata[vehicle.ID] = {}
            vehicledata[vehicle.ID]["plotdata"] = self.ev.get_periods(
                vehicle, vehicle_no, periods=period_names
            )
            vehicle_no += 1

        # Run specific tests
        self._park_inside_depot(
            vehicledata,
            name_general=periods_map["depot general"],
            name_park=periods_map["park"],
        )
        self._charge_inside_park(
            vehicledata,
            name_park=periods_map["park"],
            names_charge=periods_map["charge"],
        )
        self._service_outside_park(
            vehicledata, name_park=periods_map["park"], name_serve=periods_map["serve"]
        )
        self._trip_outside_depot(vehicledata, name_general=periods_map["depot general"])

    def _park_inside_depot(self, vehicledata, name_general, name_park):
        """Check if all parking periods are inside general depot periods."""
        result = {}
        for vID in vehicledata:
            generals = vehicledata[vID]["plotdata"][name_general]["xranges"]
            parks = vehicledata[vID]["plotdata"][name_park]["xranges"]

            found = [False] * len(parks)
            i = 0
            for park_start, park_dur in parks:
                for general_start, general_dur in generals:
                    if (
                        park_start >= general_start
                        and park_start + park_dur <= general_start + general_dur
                    ):
                        found[i] = True
                        i += 1
                        break
            result[vID] = found

        valid = True
        for ri in result.values():
            if not all(ri):
                valid = False
                break

        self.results["park_inside_depot"] = {"valid": valid, "result": result}

    def _charge_inside_park(self, vehicledata, name_park, names_charge):
        """Check if all charging periods are inside parking periods.

        names_charge relies on vehicles only having one type of charging
        process (e.g. charge_fc or charge_oc, but not both)
        """
        result = {}
        for vID in vehicledata:
            parks = vehicledata[vID]["plotdata"][name_park]["xranges"]
            # Find the applicable charging process name
            for name_charge in names_charge:
                # if 'GN' in vID:
                #     print(vID, name_charge)
                #     print(vehicledata[vID]['plotdata'])
                if vehicledata[vID]["plotdata"][name_charge]["xranges"]:
                    break
            charges = vehicledata[vID]["plotdata"][name_charge]["xranges"]

            found = [False] * len(charges)
            i = 0
            for charge_start, charge_dur in charges:
                for park_start, park_dur in parks:
                    if (
                        charge_start >= park_start
                        and charge_start + charge_dur <= park_start + park_dur
                    ):
                        found[i] = True
                        i += 1
                        break
            result[vID] = found

        valid = True
        for ri in result.values():
            if not all(ri):
                valid = False
                break

        self.results["charge_inside_park"] = {"valid": valid, "result": result}

    def _service_outside_park(self, vehicledata, name_park, name_serve):
        """Check if all service periods are outside parking periods."""
        result = {}
        for vID in vehicledata:
            parks = vehicledata[vID]["plotdata"][name_park]["xranges"]
            serves = vehicledata[vID]["plotdata"][name_serve]["xranges"]

            not_found = [True] * len(serves)
            i = 0
            for serve_start, serve_dur in serves:
                for park_start, park_dur in parks:
                    park_end = park_start + park_dur
                    if (
                        park_start < serve_start < park_end
                        or park_start < serve_start + serve_dur < park_end
                    ):
                        not_found[i] = False
                        i += 1
                        break
            result[vID] = not_found

        valid = True
        for ri in result.values():
            if not all(ri):
                valid = False
                break

        self.results["service_outside_park"] = {"valid": valid, "result": result}

    def _trip_outside_depot(self, vehicledata, name_general):
        """Check if all trip periods are outside general depot periods."""
        result = {}

        for vID in vehicledata:
            generals = vehicledata[vID]["plotdata"][name_general]["xranges"]
            trips = self.ev.vehicle_generator.select(vID).finished_trips

            not_found = [True] * len(trips)
            i = 0
            for trip in trips:
                for general_start, general_dur in generals:
                    general_end = general_start + general_dur
                    if (
                        general_start < trip.atd < general_end
                        or general_start < trip.ata < general_end
                    ):
                        not_found[i] = False
                        i += 1
                        break
            result[vID] = not_found

        valid = True
        for ri in result.values():
            if not all(ri):
                valid = False
                break

        self.results["trip_outside_depot"] = {"valid": valid, "result": result}

    def single_matches(self):
        """Check if trips are assigned only to one vehicle each."""
        trips = Counter()
        for vehicle in self.ev.vehicle_generator.items:
            trips.update(trip.ID for trip in vehicle.finished_trips)
        self.results["single_matches"] = {
            "valid": len(trips) == sum(trips.values()),
            "trips": trips,
        }



"""Script for executing the depot layout optimization with two objectives:
- max capacity (c)
- max mean of unblocked ready for departure vehicles (urfd).

Uses the DEAP evolutionary computation framework.
"""
import random
import numpy as np
from multiprocessing import Pool
import os
from deap import base, creator, tools

# Load opt settings before importing other parts of layout_opt
import eflips.depot.layout_opt.settings

filename_opt_settings = "..\\..\\..\\bus_depot\\opt_settings\\diss_kls"

eflips.depot.layout_opt.settings.load_settings(filename_opt_settings)
from eflips.depot.layout_opt.settings import OPT_CONSTANTS as OC

from eflips.depot.layout_opt import opt_tools, util, evaluation
from eflips.depot.layout_opt.opt_tools.fitness_util import memorize, lookup
import eflips.depot.layout_opt.opt_tools.fitness_c_urfd
from eflips.helperFunctions import Tictoc


# GA-parameters

# Load optimization settings
POP_SIZE = OC["algorithm"]["POP_SIZE"]

CXPB = OC["algorithm"]["CXPB"]

MUTPB = OC["algorithm"]["MUTPB"]
MUTPB_AREA_COUNT = OC["algorithm"]["MUTPB_AREA_COUNT"]
MUTSIGMA_AREA_COUNT = OC["algorithm"]["MUTSIGMA_AREA_COUNT"]
MUTPB_AREA_TYPE = OC["algorithm"]["MUTPB_AREA_TYPE"]
MUTPB_AREA_CAPACITY = OC["algorithm"]["MUTPB_AREA_CAPACITY"]
MUTSIGMA_AREA_CAPACITY = OC["algorithm"]["MUTSIGMA_AREA_CAPACITY"]

# Set max. generations
NGEN = 1000

# Set up optimization tools

creator.create("FitnessMulti", base.Fitness, weights=(1.0, 1.0))
creator.create("Individual", opt_tools.DepotPrototype, fitness=creator.FitnessMulti)

toolbox = base.Toolbox()
toolbox.register("individual", opt_tools.init_random_depot, creator.Individual)
toolbox.register("population", tools.initRepeat, list, toolbox.individual)
toolbox.register("clone", opt_tools.clone_depot)

toolbox.register("mate", opt_tools.cxOnePoint_depot)
toolbox.register(
    "mutate",
    opt_tools.mutgaussian_depot_or,
    mutsigma_area_count=MUTSIGMA_AREA_COUNT,
    mutpb_area_count=MUTPB_AREA_COUNT,
    mutpb_area_type=MUTPB_AREA_TYPE,
    mutsigma_area_capacity=MUTSIGMA_AREA_CAPACITY,
    mutpb_area_capacity=MUTPB_AREA_CAPACITY,
    splitpb=OC["algorithm"]["SPLITPB"],
)
# toolbox.register('select', tools.selTournamentDCD)
toolbox.register("select", tools.selNSGA2)
toolbox.register("evaluate", opt_tools.fitness_c_urfd.evaluate)


def main():
    """Execute the given optimization setup.
    Return the final population and statistics.
    """
    g = 0  # generation counter

    # Start time measurement
    tictoc = Tictoc(print_timestamps=False)
    tictoc.tic()

    # Initialize a population
    pop = toolbox.population(n=POP_SIZE)

    # Evaluate the entire population

    # Split individuals into uniques and duplicates to lower the evaluation
    # effort
    uniques, duplicates = util.attrbased_set(pop, "ID")
    results = list(pool.map(toolbox.evaluate, uniques))
    # results = list(map(toolbox.evaluate, uniques))
    fitnesses = [result[:-1] for result in results]
    ind_results = [result[-1] for result in results]
    for ind, fit, ind_result in zip(uniques, fitnesses, ind_results):
        ind.fitness.values = fit
        ind.results = ind_result
        memorize(ind, memory)
    for ind in duplicates:
        # Look up fitness for duplicates
        lookup(ind, memory)
    tools.emo.assignCrowdingDist(pop)

    # Prepare logging and stats
    logbook = tools.Logbook()
    logbook.header = (
        "gen",
        "evals",
        "duplicates_this_gen",
        "looked_up",
        "skipped",
        "memorized",
        "comptime",
        "fitness",
    )
    logbook.chapters["fitness"].header = "min", "avg", "max", "std", "feasible"
    cx_this_gen = 0
    crossovers = []
    # Log stats
    record = mstats.compile(pop)
    skipped = sum(ind.results["simtime"] == 0 for ind in pop)
    feasible = sum(
        eflips.depot.layout_opt.opt_tools.fitness_c_urfd.feasible(ind) for ind in pop
    ) / len(pop)
    logbook.record(
        gen=g,
        evals=len(pop),
        duplicates_this_gen=len(duplicates),
        looked_up=0,
        skipped=skipped,
        memorized=len(memory),
        comptime=tictoc.last_interval,
        feasible=feasible,
        **record
    )
    print()
    print(logbook.stream)  # values of initial population
    hof.update(pop)
    tictoc.toc()

    while not stopping_criteria.check(
        g=g,
        ngen=NGEN,
        hof=hof,
        maxfitness=maxfitness_estimate,
        improvement_interval=30,
        pop=pop,
        feasible=eflips.depot.layout_opt.opt_tools.fitness_c_urfd.feasible,
    ):
        g += 1

        # Clone the population to prepare the offspring
        offspring = list(pool.map(toolbox.clone, pop))
        # offspring = list(map(toolbox.clone, pop))

        # Apply crossover on the offspring
        for child1, child2 in zip(offspring[::2], offspring[1::2]):
            if random.random() < CXPB:
                toolbox.mate(child1, child2)
                del child1.fitness.values
                del child2.fitness.values
                cx_this_gen += 1

        # Apply mutation on the offspring
        for i in range(2):
            for mutant in offspring:
                if random.random() < MUTPB:
                    mutated = toolbox.mutate(mutant)
                    if mutated:
                        del mutant.fitness.values

        # Evaluate the fitness of the offspring

        modified_inds = [ind for ind in offspring if not ind.fitness.valid]
        # Split modified individuals into uniques and duplicates to lower the
        # evaluation effort
        uniques, duplicates = util.attrbased_set(modified_inds, "ID")
        # Look up fitness for individuals that are memorized
        for ind in uniques:
            lookup(ind, memory)
        # Evaluate the individuals that really are new
        new_inds = [ind for ind in uniques if not ind.fitness.valid]
        results = list(pool.map(toolbox.evaluate, new_inds))
        # results = list(map(toolbox.evaluate, new_inds))
        fitnesses = [result[:-1] for result in results]
        ind_results = [result[-1] for result in results]
        for ind, fit, sim_result in zip(new_inds, fitnesses, ind_results):
            ind.fitness.values = fit
            ind.results = sim_result
            memorize(ind, memory)
        # Look up fitness for individuals that are duplicate in this generation
        # or memorized even earlier
        for ind in duplicates:
            lookup(ind, memory)

        # Select the next generation individuals from pop and offspring
        tools.emo.assignCrowdingDist(pop + offspring)  # required for selTournamentDCD
        pop[:] = toolbox.select(pop + offspring, POP_SIZE)

        # Log stats
        hof.update(pop)
        crossovers.append(cx_this_gen)
        cx_this_gen = 0
        record = mstats.compile(pop)
        skipped = sum(ind.results["simtime"] == 0 for ind in new_inds)
        feasible = sum(
            eflips.depot.layout_opt.opt_tools.fitness_c_urfd.feasible(ind)
            for ind in pop
        ) / len(pop)
        logbook.record(
            gen=g,
            evals=len(new_inds),
            duplicates_this_gen=len(duplicates),
            looked_up=len(modified_inds) - len(new_inds),
            skipped=skipped,
            memorized=len(memory),
            comptime=tictoc.last_interval,
            feasible=feasible,
            **record
        )
        print(logbook.stream)
        tictoc.toc()

    tictoc.print_timestamps = True
    tictoc.toc()
    print("(%f s per toc)" % (tictoc.tlist[-1] / tictoc.nTocs - 1))
    print()
    return pop, logbook, crossovers


if __name__ == "__main__":
    opt_tools.init.print_area_precomps()

    # Set up multiprocessing (must be protected by "if __name__ == '__main__'")
    n_processes = os.cpu_count() - 1  # os.cpu_count() is default
    pool = Pool(n_processes, maxtasksperchild=100)
    # maxtasksperchild: renew processes after a certain amount of tasks (proved
    # to be critical for long optimization runs)
    print("Multiprocessing with %d processes" % n_processes)
    toolbox.register("map", pool.map)

    # Prepare statistics
    fitness_stats = tools.Statistics(key=lambda ind: ind.fitness.values)
    mstats = tools.MultiStatistics(fitness=fitness_stats)
    mstats.register("avg", np.mean, axis=0)
    mstats.register("std", np.std, axis=0)
    mstats.register("min", np.min, axis=0)
    mstats.register("max", np.max, axis=0)

    hof = tools.ParetoFront()

    # Set up memory and other logging
    memory = {}

    # Set stopping criteria
    maxfitness_estimate = creator.FitnessMulti()
    maxfitness_estimate.values = (opt_tools.init.CAPACITY_MAX, 102)
    stopping_criteria = util.StoppingCriteria()
    # stopping_criteria.select('max_gen_reached', 'max_fit_reached', 'no_improvement')
    # stopping_criteria.select('no_improvement')
    stopping_criteria.select("max_gen_reached")
    # stopping_criteria.select('max_gen_reached', 'no_improvement')
    # stopping_criteria.select('max_fit_reached')
    # stopping_criteria.select('max_fit_reached', 'no_improvement')
    # stopping_criteria.select('feasible_found')
    # stopping_criteria.select('feasible_found', 'max_gen_reached')

    # Run optimization
    print("Estimated max fitness: %s" % maxfitness_estimate)
    pop, logbook, crossovers = main()

    # Gather data and do some evaluation
    ev = evaluation.OptEvaluation(
        pop,
        logbook,
        hof,
        crossovers,
        eflips.depot.layout_opt.opt_tools.mutation.mutations,
        memory,
        eflips.depot.layout_opt.opt_tools.fitness_c_urfd.feasible,
        eflips.depot.layout_opt.opt_tools.fitness_c_urfd.feasible_fr_vec,
    )
    ev.results_operators()
    ev.results_feasbility()
    ev.results_simtimes()

    # ev.save()



# -*- coding: utf-8 -*-
"""
Components to model multi criteria decision problems. Application for park- and
dispatch strategies.

"""
import numpy as np
from abc import ABC, abstractmethod
import eflips


# Settings

# For ParkRating
rfd_diff_park_lower_bound = -1200  # s
rfd_diff_park_upper_bound = 7200  # s
park_rating_weights = {
    "buffer": 22 / 72,  # PM 7/22
    "typestack": 13 / 72,  # PM 2/11
    "rfd_diff_pos": 8 / 72,  # PM 7/66
    "rfd_diff_neg": 29 / 144,  # PM 5/22
    "available_power": 2 / 72,  # PM 1/66
    "empty_slots_exit": 25 / 144,  # PM5/33
}

# For DispatchRating
rfd_diff_dispatch_upper_bound = 10800  # s
dispatch_rating_weights = {
    "buffer": 10 / 50,  # PM 1/5
    "typestack": 16 / 50,  # PM 17/45
    "rfd_diff": 25 / 100,  # PM 13/45
    "available_power": 3 / 50,  # PM 1/45
    "empty_slots_exit": 17 / 100,  # PM 1/9
}


class Rating:
    """Data container with methods to solve multi criteria decision problems.

    Parameters:
    alternatives: [list or 2D numpy.ndarray] of lists that represent an
        alternative. Alternatives must contain numerical data and have the same
        length. Is converted to a 2D numpy.ndarray during init.
    weights: [list or 1D numpy.ndarray or None] of factors for weighting values
        of an alternative. Is converted to a 1D numpy.ndarray during init.

    Attributes:
    best_value: [int or float] the best value after solving.
    best_alternative_nos: [tuple] from np.where with the row indices of the
        best alternatives in alternatives. E.g.: (array([2], dtype=int64),)
    best_alternatives: [list] of the best alternatives.

    """

    def __init__(self, alternatives, weights=None):
        self.alternatives, self.weights = self.check_input(alternatives, weights)
        self.weighted = None
        self.sums = None
        self.best_value = None
        self.best_alternative_nos = None
        self.best_alternatives = None

    @staticmethod
    def check_input(alternatives, weights):
        """Check alternatives and weights for validity."""
        if not isinstance(alternatives, list) or not all(
            isinstance(alt, list) for alt in alternatives
        ):
            raise ValueError(
                "Rating parameter 'alternatives' must be a list " "of lists."
            )
        if not alternatives:
            raise ValueError("Rating parameter 'alternatives' cannot be " "empty.")
        len_first = len(alternatives[0])
        if not all(len(i) == len_first for i in alternatives):
            raise ValueError(
                "Entries in Rating parameter 'alternatives' must"
                " have the same length."
            )
        if weights is not None:
            if not isinstance(weights, list) or not len(weights) == len_first:
                raise ValueError(
                    "Rating parameter 'weights' must be a list with the same "
                    "length as entries in 'alternatives'."
                )
        return np.asarray(alternatives), np.asarray(weights)

    def weighted_sum(self):
        """Locate the maximum weighted sums in self.alternatives."""
        if self.weights is None:
            raise ValueError("weights must be supplied for weighted_sum")

        self.weighted = self.weights * self.alternatives
        self.sums = self.weighted.sum(axis=1)
        self.best_value = self.sums.max()
        self.best_alternative_nos = np.where(self.sums == self.best_value)
        self.best_alternatives = self.alternatives[self.best_alternative_nos].tolist()


class BaseCriterion(ABC):
    """Base class for a criterion.
    Subclasses must implement attibute 'value' and method 'calculate'.

    """

    @abstractmethod
    def calculate(self, *args, **kwargs):
        """Determine self.value from inputs."""


class BufferPark(BaseCriterion):
    """Criterion for ParkRating.
    Exclusive for Direct areas. Value is 0 for Line.

    """

    def __init__(self, area, vehicle):
        self.value = self.calculate(area, vehicle)

    def calculate(self, area, vehicle):
        if not isinstance(area, eflips.depot.DirectArea):
            return 0

        # Check if Direct areas are all emtpy
        count_direct = sum(
            sum(a.count for a in g.direct_areas) for g in area.depot.parking_area_groups
        )
        if count_direct == 0:
            return 1

        total_buffer_capacity = area.depot.parking_capacity_direct

        if vehicle.vehicle_type.group:
            count = sum(
                sum(
                    v.vehicle_type.group is vehicle.vehicle_type.group
                    for v in a.items
                    if v
                )
                for a in area.parking_area_group.direct_areas
            )
            share_target = vehicle.vehicle_type.group.share[area.depot]
            # print('group. count: %d, share_target: %f' % (count, share_target))
        else:
            count = sum(
                sum(v.vehicle_type is vehicle.vehicle_type for v in a.items if v)
                for a in area.parking_area_group.direct_areas
            )
            share_target = vehicle.vehicle_type.share[area.depot]
            # print('lone. count: %d, share_target: %f' % (count, share_target))

        share = count / total_buffer_capacity

        result = share < share_target
        # print('BufferPark for vehicle %s: count=%s, total=%s, share=%s, share_target=%s'
        #       % (vehicle.ID, count, total_buffer_capacity, share, share_target))
        if result:
            return 1
        else:
            return 0


class TypestackPark(BaseCriterion):
    """Criterion for ParkRating.
    Exclusive for Line areas. Value is 0 for Direct.

    """

    def __init__(self, area, vehicle):
        self.value = self.calculate(area, vehicle)

    def calculate(self, area, vehicle):
        if not isinstance(area, eflips.depot.LineArea):
            return 0

        result = area.istypestack_with(vehicle)
        if result is None:
            return 0
        elif result:
            return 1
        else:
            return -1


class RfdDiffPark(BaseCriterion):
    """Criterion for ParkRating.
    Exclusive for Line areas where *vehicle* would be blocked. Value is 0 for
    Direct.

    slot: [tuple] with items (area, index of slot)

    """

    lower_bound = rfd_diff_park_lower_bound
    upper_bound = rfd_diff_park_upper_bound

    def __init__(self, slot, vehicle):
        self.value, self.diff = self.calculate(slot, vehicle)

    def calculate(self, slot, vehicle):
        area, index = slot[0], slot[1]

        if not isinstance(area, eflips.depot.LineArea):
            return 0, None

        # Get the blocking vehicle, if there is one
        side_get = area.side_get_default
        index_blocking = area.index_neighbour(index, side_get)
        if index_blocking is None:
            return 0, None
        blocking_vehicle = area.items[index_blocking]

        # Get the rfd-diff
        charge_proc = area.charge_proc  # type of charging process at the slot
        dur_est_blocked = charge_proc.estimate_duration(
            vehicle, slot[0].charging_interfaces[slot[1]]
        )
        etc_blocking = blocking_vehicle.dwd.etc_processes
        diff = 0
        if isinstance(etc_blocking, int):
            diff = slot[0].env.now + dur_est_blocked - etc_blocking
        elif etc_blocking is eflips.depot.EstimateValue.UNKNOWN:
            diff = dur_est_blocked - charge_proc.estimate_duration(
                blocking_vehicle, slot[0].charging_interfaces[index_blocking]
            )
        elif etc_blocking is eflips.depot.EstimateValue.COMPLETED:
            diff = dur_est_blocked

        # Determine value based on rfd-diff
        if diff < self.lower_bound:
            return -1, diff
        elif self.lower_bound <= diff < 0:
            return (1 / 1200) * diff, diff
        elif 0 <= diff < self.upper_bound:
            return -(1 / 3600) * diff + 1, diff
        elif self.upper_bound <= diff:
            return -1, diff


class AvailablePower(BaseCriterion):
    """Criterion for ParkRating and DispatchRating.
    For both Direct and Line areas.

    slot: [tuple] with items (area, index of slot)

    """

    def __init__(self, slot, max_power):
        self.value = self.calculate(slot, max_power)

    def calculate(self, slot, max_power):
        normalization_factor = 1 / max_power
        return slot[0].charging_interfaces[slot[1]].max_power * normalization_factor


class EmptySlotsExitPark(BaseCriterion):
    """Criterion for ParkRating.
    Exclusive for Line areas. Value is 0 for Direct.

    """

    def __init__(self, area, max_capacity_line):
        self.value = self.calculate(area, max_capacity_line)

    def calculate(self, area, max_capacity_line):
        if not isinstance(area, eflips.depot.LineArea):
            return 0

        if area.capacity == 2:
            # Can't have a blocked emtpy slot at the exit and another unblocked
            # empty slot if capacity=2
            return 0

        normalization_factor = 1 / (max_capacity_line - 2)
        return -area.vacant_blocked * normalization_factor


class SlotAlternative:
    """Summary of criteria for one slot for ParkRating.

    slot: [tuple] with items (area, index of slot)
    vehicle: [SimpleVehicle]
    max_power: [int] maximum power in kW of all charging interfaces at parking
        areas
    max_capacity_line: [int or None] max capacity of Line parking areas

    """

    def __init__(self, slot, vehicle, max_power, max_capacity_line):
        self.slot = slot
        self.vehicle = vehicle

        # Determine criteria
        self.buffer = BufferPark(slot[0], vehicle)
        self.typestack = TypestackPark(slot[0], vehicle)
        self.rfd_diff = RfdDiffPark(slot, vehicle)
        self.available_power = AvailablePower(slot, max_power)
        self.empty_slots_exit = EmptySlotsExitPark(slot[0], max_capacity_line)

        # Split rfd_diff value into pos and neg
        if self.rfd_diff.diff is None:
            self.rfd_diff_pos = 0
            self.rfd_diff_neg = 0
        elif self.rfd_diff.diff >= 0:
            self.rfd_diff_pos = self.rfd_diff.value
            self.rfd_diff_neg = 0
        else:
            self.rfd_diff_pos = 0
            self.rfd_diff_neg = self.rfd_diff.value

        # Get values suitable for rating
        self.values = [
            self.buffer.value,
            self.typestack.value,
            self.rfd_diff_pos,
            self.rfd_diff_neg,
            self.available_power.value,
            self.empty_slots_exit.value,
        ]


class ParkRating(Rating):
    """

    alternatives_obj: [list] of SlotAlternative instances

    """

    weights = [
        park_rating_weights["buffer"],
        park_rating_weights["typestack"],
        park_rating_weights["rfd_diff_pos"],
        park_rating_weights["rfd_diff_neg"],
        park_rating_weights["available_power"],
        park_rating_weights["empty_slots_exit"],
    ]

    def __init__(self, alternatives_obj):
        self.alternatives_obj = alternatives_obj

        super(ParkRating, self).__init__(
            [alt.values for alt in alternatives_obj], self.weights
        )
        self.weighted_sum()


class BufferDispatch(BaseCriterion):
    """Criterion for DispatchRating.
    Exclusive for Direct areas. Value is 0 for Line.

    """

    def __init__(self, area, vehicle):
        self.value = self.calculate(area, vehicle)

    def calculate(self, area, vehicle):
        if not isinstance(area, eflips.depot.DirectArea):
            return 0
        else:
            return -1


class TypestackDispatch(BaseCriterion):
    """Criterion for DispatchRating.
    Exclusive for Line areas. Value is 0 for Direct.

    """

    def __init__(self, area):
        self.value = self.calculate(area)

    def calculate(self, area):
        if not isinstance(area, eflips.depot.LineArea):
            return 0

        result, vtype = area.istypestack()
        if result is None:
            raise RuntimeError(
                "Area %s cannot be empty when applying the"
                "TypestackDispatch criterion." % area.ID
            )
        elif result:
            # typestack
            return 0
        else:
            # mixed types
            return 1


class RfdDiffDispatch:
    """Criterion for DispatchRating.
    Exclusive for Line areas. Value is 0 for Direct.

    slot: [tuple] with items (area, index of slot). Must contain a vehicle that
        is ready for departure, i.e. dwd.etc_processes returns
        eflips.EstimateValue.COMPLETED.

    """

    upper_bound = rfd_diff_dispatch_upper_bound

    def __init__(self, slot):
        self.value, self.diff = self.calculate(slot)

    def calculate(self, slot):
        area, index = slot[0], slot[1]
        if not isinstance(area, eflips.depot.LineArea):
            return 0, None

        max_diff = self.get_max_diff(slot)
        # Determine value based on rfd-diff
        if 0 <= max_diff < self.upper_bound:
            return -(1 / 10800) * max_diff + 1, max_diff
        elif max_diff >= self.upper_bound:
            return 0, max_diff
        elif max_diff < 0:
            # Negative diff is impossible if the blocking vehicle is rfd. Yet,
            # return 1 in this case to enable considering vehicles that are not
            # rfd.
            return 1, max_diff

    @staticmethod
    def get_max_diff(slot):
        """Get the maximum of all etcs (estimated time of completion) of all
        vehicles at the area of *slot* and return it as [int].

        slot: [tuple] with items (area, index of slot)
        """
        area = slot[0]
        charge_proc = area.charge_proc  # type of charging process at the slot
        diffs = []
        for vehicle in area.vehicles:
            etc = vehicle.dwd.etc_processes
            if isinstance(etc, int):
                diff = etc - slot[0].env.now
                diffs.append(diff)
            # Convert EstimateValue to numerical rfddiff
            elif etc is eflips.depot.EstimateValue.UNKNOWN:
                diff = charge_proc.estimate_duration(
                    vehicle, slot[0].charging_interfaces[area.items.index(vehicle)]
                )
                diffs.append(diff)
            elif etc is eflips.depot.EstimateValue.COMPLETED:
                diff = 0
                diffs.append(diff)

        return max(diffs)


class EmptySlotsExitDispatch(BaseCriterion):
    """Criterion for DispatchRating.
    Exclusive for Line areas. Value is 0 for Direct.

    """

    def __init__(self, area, max_capacity_line):
        self.value = self.calculate(area, max_capacity_line)

    def calculate(self, area, max_capacity_line):
        if not isinstance(area, eflips.depot.LineArea):
            return 0

        normalization_factor = 1 / (max_capacity_line - 1)
        return area.vacant_blocked * normalization_factor


class VehicleAlternative:
    """Summary of criteria for one vehicle for DispatchRating.

    slot: [tuple] with items (area, index of vehicle's slot)
    vehicle: [SimpleVehicle]
    max_power: [int] maximum power in kW of all charging interfaces at parking
        areas
    max_capacity_line: [int or None] max capacity of Line parking areas

    """

    def __init__(self, slot, vehicle, max_power, max_capacity_line):
        self.slot = slot
        self.vehicle = vehicle

        # Determine criteria
        self.buffer = BufferDispatch(slot[0], vehicle)
        self.typestack = TypestackDispatch(slot[0])
        self.rfd_diff = RfdDiffDispatch(slot)
        self.available_power = AvailablePower(slot, max_power)
        self.empty_slots_exit = EmptySlotsExitDispatch(slot[0], max_capacity_line)

        # Get values suitable for rating
        self.values = [
            self.buffer.value,
            self.typestack.value,
            self.rfd_diff.value,
            self.available_power.value,
            self.empty_slots_exit.value,
        ]


class DispatchRating(Rating):
    """

    alternatives_obj: [list] of VehicleAlternative instances

    """

    weights = [
        dispatch_rating_weights["buffer"],
        dispatch_rating_weights["typestack"],
        dispatch_rating_weights["rfd_diff"],
        dispatch_rating_weights["available_power"],
        dispatch_rating_weights["empty_slots_exit"],
    ]

    def __init__(self, alternatives_obj):
        self.alternatives_obj = alternatives_obj

        super(DispatchRating, self).__init__(
            [alt.values for alt in alternatives_obj], self.weights
        )
        self.weighted_sum()
